/**********************************************************************************
21/November/2000 (c) SCEE. Written by Jason Page.

main.c
	Initialises everything, main loop, controller input, graphics...all the usual stuff..
***********************************************************************************/


#define BOOT_HOST 0
//#define BOOT_CD
#include <eekernel.h>
#include <malloc.h>
#include <stdio.h>
#include <libsdr.h>
#include <sdrcmd.h>
#include <libdev.h>
#include <libpad.h>
#include <libgraph.h>
#include <libpkt.h>
#include <libcdvd.h>
#include <sifdev.h>
#include <sif.h>
#include <sifcmd.h>
#include "xmrpc.h"
#include "../iop/xmcalls.h"

#include "graphics.h"

#define SUNIT		0x01
#define PACKETSIZE	(0x100*SUNIT)
#define WORKSIZE	(0x80)
#define WORKBASE	0x70000000
#define SUPPER		(WORKBASE+WORKSIZE)
#define SLOWER		(WORKBASE+WORKSIZE+PACKETSIZE)
#define SCREEN_WIDTH	640
#define SCREEN_HEIGHT	224

sceDmaChan *dmaGIF=NULL;	// DMA channel for string

#define CNTL_RUN  0
#define CNTL_STOP 1

void ReverbTest(void);

// Filenames (Thanks to Skaven for letting me use these tunes...)
#define XM_FILENAME "host0:revenge.xm"
#define SAMPLEBANK_FILENAME "host0:revenge.xmb"
#define SAMPLEHEADER_FILENAME "host0:revenge.xmh"

#define XM_FILENAME2 "host0:icefront.xm"
#define SAMPLEBANK_FILENAME2 "host0:icefront.xmb"
#define SAMPLEHEADER_FILENAME2 "host0:icefront.xmh"


u_char *helptext[]={
	"(Left = Move backwards through song)\n",
	"(Right = Move forwards through song)\n",
	"(Up = Close song / Close XM File)\n",
	"(Down = Open song / Open XM File)\n",
	"(X = Play song)\n",
	"(Triangle = Pause / Resume song)\n",
	"(Circle = Stop song)\n",
	"(L2 = Song 1)\n",
	"(R2 = Song 2)\n",
	"(Start = Exit program)\n"
	};

u_char *songtext[]={
	"> Revenge - By Peter 'Skaven' Hajba.\n",
	"> IceFront - By Peter 'Skaven' Hajba.\n"
	};

int HelpTextCounter=0;
int HelpTextPos=0;

int XM_ID[2]={0,1};				// ID's that will be associated with each XM File (range from 0-7)
int XM_SONG_ID[2]={0,1};		// ID's that will be associated with each song (max 31)
int XM_SAMPLEBANK_ID[2]={0,1};	// ID's that will be associated with each sample bank
int XM_IOPAddress[2];			// Address of XM file in IOP RAM
int FirstSPUChannel[2]={0,12};	// First SPU channel to playback song (first song therefore uses 12 channels...)
u_int SAMPLEBANK_ADDR[2],SAMPLEHEADER_ADDR[2],XM_ADDR[2];		// Holds the files EE address once loaded
u_int SAMPLEBANK_SIZE[2],SAMPLEHEADER_SIZE[2],XM_SIZE[2];		// Holds the files sizes
int Buffer64K;	// IOP Address of allocated 64K buffer (where samples are transfered from EE to IOP)
int CS=0;		// Current Song to play (used for showing info/user control..)


static int Gs_Initialize(void);
void DisplayText(void);
int GetUserControl(void);
void SetCursorValue(int command);
static int console;
extern void sceDevConsLocate(int cd, u_int lx, u_int ly);
static sceGsDBuff db;
int LoadFiles(void);
int TestTransfer(void);

int paddata, oldpad;
u_char rdata[32];

int LastVolumes[32];	// For volume display bars..

int frame=0;			// Frame Counter

int CursorX=320;
int CursorY=100;
int CellX,CellY;
int PadGate=0;


// JoyPad buffers
u_long128 pad_dma_buf[scePadDmaBufferMax] __attribute__((aligned (64)));
u_long128 pad_dma_buf2[scePadDmaBufferMax] __attribute__((aligned (64)));

/*------------------------------------------
load_modules
	Loads required sif modules to the iop.
  ------------------------------------------*/
#define _MOD_ROOT    "host0:/usr/local/sce/iop/modules/"
#define _MOD_SIO2MAN "sio2man.irx"
#define _MOD_PADMAN  "padman.irx"
#define _MOD_LIBSD   "libsd.irx"
#define _MOD_XMPLAY "host0:../iop/xmplay.irx"

/*****************************************************************************
load_modules
	load and transfer relevent IRX modules to IOP
*****************************************************************************/
void load_modules (void)
{
int i;
	for (i=0;i<32;i++)
	{
		LastVolumes[i]=0;
	}

    sceSifInitRpc (0);

    printf ("loading modules ....");

    // Padp
    if (sceSifLoadModule (_MOD_ROOT _MOD_SIO2MAN, 0, NULL) < 0)
    {
		printf ("  loading %s failed\n", _MOD_SIO2MAN);
		exit (0);
    }
    if (sceSifLoadModule (_MOD_ROOT _MOD_PADMAN, 0, NULL) < 0)
    {
		printf ("  loading %s failed\n", _MOD_PADMAN);
		exit (0);
    }
    // libsd
    if (sceSifLoadModule( _MOD_ROOT _MOD_LIBSD, 0, NULL) < 0)
    {
		printf("  loading libsd.irx failed\n");
		exit (0);
    }

	// LOAD XMPlayer module
    if (sceSifLoadModule (_MOD_XMPLAY, 0, NULL) < 0)
	{
		printf("  loading xmplay.irx failed\n");
		exit(0);
    }
    printf ("Complete.\n");
    return;
}

/*****************************************************************************
init_xm
	Initialise XM IOP module. Init SPU (set both cores to full volume).
	Call TestTransfer (load XM files, initialise them, play song)
*****************************************************************************/
int init_xm (void)
{
    if (SOUND_InitIOP () < 0)
		return (-1);

	SOUND_InitSPU();		// Init IOP LibSD & Set master volumes for both cores
	FlushIOPCommand(0);		// Wait for command to be processed on IOP
	TestTransfer();
    return 0;
}

/*****************************************************************************
main_loop
	init graphics. init joypad.
	loop displays graphics, gets info for XM song.
*****************************************************************************/
int main_loop(void)
{
int QuitFlag;
	Gs_Initialize();
	JInitGraphics();
    dmaGIF = sceDmaGetChan( SCE_DMA_GIF );

	sceDevConsInit();
	console = sceDevConsOpen(OFFX + (16<<4), OFFY + (24<<4), 75, 30);
	sceDevConsClear(console);

	scePadInit(0);
	scePadPortOpen(0, 0, pad_dma_buf);

	SetAlphaBlend();

    while( sceGsSyncV(0) == 0 );

	frame=0;

	while( 1 )
	{

		QuitFlag=GetUserControl();
		if (QuitFlag!=0)
		{
			XMEE_QuickShutdown(XM_ID[0],XM_SONG_ID[0]);	// Stops song, Free's Song,XMBuffer and XMPattern IOP RAM.
			XMEE_QuickShutdown(XM_ID[1],XM_SONG_ID[1]);	// Stops song, Free's Song,XMBuffer and XMPattern IOP RAM.
			free((void*)XM_ADDR[0]);
			free((void*)XM_ADDR[1]);
			return;
		}

		FlushCache(0);
		DisplayText();

		HelpTextCounter++;
		if (HelpTextCounter>100)
		{
			HelpTextCounter=0;
			HelpTextPos++;
			HelpTextPos%=10;
		}

		XM_GetSongFeedback(CS);			// songid
		FlushIOPCommand(SIF_RPCM_NOWAIT);						// transfer iop buffer
		XMEE_SetFeedback();


		if(frame&0x01)
		{
			sceGsSetHalfOffset(&db.draw1,2048,2048, sceGsSyncV(0)^0x01);
		}
		else
		{
			sceGsSetHalfOffset(&db.draw0,2048,2048, sceGsSyncV(0)^0x01);
		}
		sceGsSwapDBuff(&db, frame);
		frame++;
		FlushCache(0);
		sceGsSyncPath(0,0);
	}
	return;
}

/*****************************************************************************
main
	calls everything.
*****************************************************************************/
int main( int argc, char *argv[]  )
{
int a;

	sceSifInitRpc(0);
    load_modules ();

    printf ("XMPLAY Initialize ....\n");
    a=init_xm();
    printf ("XMPLAY MainLoop ....\n");
	if (a==0)
	{
		Gs_Initialize();
	    main_loop ();
	}
	printf("XMPLAY Finished...\n\n\n");
	return 0;
}

/*****************************************************************************
Gs_Initialize
	graphics stuff. Don't know - don't care.
*****************************************************************************/
static int Gs_Initialize(void)
{
	int sindex;
	sceVif1Packet packet[2];
	sceDmaEnv env;
	sceDmaChan *p1;
	u_long giftagAD[2] = { SCE_GIF_SET_TAG(0, 1, 0, 0, 0, 1),
			 0x000000000000000eL };
	sceDevVif0Reset();
	sceDevVu0Reset();
	sceGsResetPath();
	sceDmaReset(1);

	sceVif1PkInit(&packet[0], (u_long128 *)SUPPER);
	sceVif1PkInit(&packet[1], (u_long128 *)SLOWER);

	sceDmaGetEnv(&env);
	env.notify = 1<<SCE_DMA_VIF1; // notify channel 
	sceDmaPutEnv(&env);

	p1 = sceDmaGetChan(SCE_DMA_VIF1);
	p1->chcr.TTE = 1;

	sceGsResetGraph(0, SCE_GS_INTERLACE, SCE_GS_NTSC, SCE_GS_FRAME);
	sceGsSetDefDBuff(&db, SCE_GS_PSMCT32, SCREEN_WIDTH, SCREEN_HEIGHT,
		SCE_GS_ZGEQUAL, SCE_GS_PSMZ24, SCE_GS_CLEAR);

	db.clear0.rgbaq.R = 0x00;
	db.clear0.rgbaq.G = 0x00;
	db.clear0.rgbaq.B = 0x00;

	db.clear1.rgbaq.R = 0x00;
	db.clear1.rgbaq.G = 0x00;
	db.clear1.rgbaq.B = 0x00;

	FlushCache(0);

	sindex = 0;
	sceVif1PkReset(&packet[sindex]);
	sceVif1PkCnt(&packet[sindex], 0);
	sceVif1PkOpenDirectCode(&packet[sindex], 0);
	sceVif1PkOpenGifTag(&packet[sindex], *(u_long128*)&giftagAD[0]);

	sceVif1PkReserve(&packet[sindex],
		sceGsSetDefAlphaEnv((sceGsAlphaEnv *)packet[sindex].pCurrent,0)
		* 4);
	sceVif1PkCloseGifTag(&packet[sindex]);
	sceVif1PkCloseDirectCode(&packet[sindex]);
	sceVif1PkEnd(&packet[sindex], 0);
	sceVif1PkTerminate(&packet[sindex]);

	// kick Gs initialize packet 
	sceDmaSend(p1,(u_int *)
		(((u_int)packet[sindex].pBase&0x0fffffff)|0x80000000));

	// display next in odd field when interlace 
	while(!sceGsSyncV(0));
}

/*****************************************************************************
DisplayText
	Displays both text and sprites
*****************************************************************************/
void DisplayText(void)
{
int a,b,i,ch,a2;
unsigned long total1;
u_char *filename;

	sceDevConsClear(console);
	sceDevConsLocate(console, 0, 0);
	sceDevConsPrintf(console, "XM PLAYER 23rd November 2000 (JasonPage/SCEE)\n");
	sceDevConsPrintf(console, "%s\n\n",helptext[HelpTextPos]);

	sceDevConsPrintf(console, "Feedback for song id %x ..%s\n\n",CS,songtext[CS]);



	sceDevConsPrintf(console, "Song Status: ");
	if (XM_SongStatus==XM_STOPPED)
		sceDevConsPrintf(console, "STOPPED (Press X to play)\n");
	else if (XM_SongStatus==XM_PLAYING)
		sceDevConsPrintf(console, "PLAYING\n");
	else if (XM_SongStatus==XM_PAUSED)
		sceDevConsPrintf(console, "PAUSED (Press triangle)\n");
	else if (XM_SongStatus==XM_NO_SONG_ALLOCATED)
		sceDevConsPrintf(console, "NO SONG ALLOCATED (Press down button)\n");
	else if (XM_SongStatus==XM_NOT_INIT)
		sceDevConsPrintf(console, "INITIALISED BUT NOT PLAYING (Press X to play)\n");
	else
		sceDevConsPrintf(console, ">>> UNKNOWN STATUS (ERROR?) <<<\n");

	if ((XM_SongStatus!=XM_NO_SONG_ALLOCATED)&&(XM_SongStatus!=XM_NOT_INIT))
	{
		sceDevConsPrintf(console, "Song Length: %d\n",XM_SongLength);
		sceDevConsPrintf(console, "Song Channels used: %d\n",XM_SongChannels);
		sceDevConsPrintf(console, "Song Position: %d\n",XM_SongPos);
		sceDevConsPrintf(console, "Song BPM: %d\n",XM_SongBPM);
		sceDevConsPrintf(console, "Song Speed: %d\n",XM_SongSpeed);
		sceDevConsPrintf(console, "Song Current Pattern: 0x%x\n",XM_CurrentPattern);
		sceDevConsPrintf(console, "Song Pattern: Pos %d\n",XM_SongPatternPos);

		a=XM_SongChannels*16;
		a/=2;
		a-=XM_SongLength;
	    SetSprite(64+a,204,XM_SongLength*2,7,64,0,100);
	    SetSprite(64+a,204,XM_SongPos*2,7,0,200,64);
	}

	sceDevConsDraw(console);


	for (i=0;i<XM_SongChannels;i++)
	{
		if (XM_SongChannelVolumes[i]>LastVolumes[i])
			LastVolumes[i]=XM_SongChannelVolumes[i];
		a=LastVolumes[i];
	    SetSprite(64+(i*16),200,14,-1+(-(a/2)),128,a*4,a*2);
		if (LastVolumes[i]>0)
			LastVolumes[i]-=2;
	}

    SetSprite(CursorX-2,CursorY-2,8,4,frame,255,frame/8);
	DrawSprites();
}

/*****************************************************************************
GetUserControl
	Reads joypad
*****************************************************************************/
int GetUserControl(void)
{
int menuspd;
int a,b,pg;

	pg=0;
    if(scePadRead(0, 0, rdata) > 0)
    {

// Left analog joystick...(not really used in this demo...!)
		paddata = (rdata[6]);				// left joystick X
		if ((paddata>0x60)&&(paddata<0xa0))
			paddata=0x80;
		paddata-=0x80;
		CursorX+=paddata/16;				// Modify Cursor X position
		paddata = (rdata[7]);				// left joystick Y
		if ((paddata>0x60)&&(paddata<0xa0))
			paddata=0x80;
		paddata-=0x80;
		CursorY+=paddata/32;				// Modify Cursor Y position
		if (CursorX<8)						// Limit cursor X position
			CursorX=8;
		else if (CursorX>638)
			CursorX=638;
		if (CursorY<8)						// Limit cursor Y position
			CursorY=8;
		else if (CursorY>220)
			CursorY=220;
// Set cursor position...
		CellX=CursorX/16;					// Set Cell values, used for checking menu positions.
		CellY=CursorY/8;

		paddata = 0xffff ^ ((rdata[2] << 8) | rdata[3]);

		if ((paddata & SCE_PADRup)&&(!(oldpad&SCE_PADRup)))
		{
			if (XM_SongStatus==XM_PLAYING)	// Pause / Resume song
				XM_Pause(CS);
			else if (XM_SongStatus==XM_PAUSED)
				XM_Resume(CS);
		}

		if ((paddata & SCE_PADLright)&&(!(oldpad&SCE_PADLright) ))
		{
			XM_SetSongPos(CS,XM_SongPos+1);	// Move forwards through song
		}
		if ((paddata & SCE_PADLleft)&&(!(oldpad&SCE_PADLleft) ))
		{
			XM_SetSongPos(CS,XM_SongPos-1);	// Move backwards through song
		}

		if (paddata & SCE_PADRright)
		{
			XM_XMQuit(CS);		// Stop song
		}

		if (paddata & SCE_PADstart )	// Press START to EXIT
		{
			return(-1);
		}

		if ((paddata & SCE_PADL2)&&(!(oldpad&SCE_PADL2)))
			CS=0;		// Set user controls / feedback to song 0
		if ((paddata & SCE_PADR2)&&(!(oldpad&SCE_PADR2)))
			CS=1;		// Set user controls / feedback to song 1

// Play the song (do this after the full init..)
		if ((paddata & SCE_PADRdown)&&(!(oldpad&SCE_PADRdown) ))
			XM_XMPlay(XM_SAMPLEBANK_ID[CS],XM_ID[CS],XM_SONG_ID[CS],FirstSPUChannel[CS],XM_Loop,-1,0);

// Perform full shutdown of XM data (song and XM File)
		if ((paddata & SCE_PADLup)&&(!(oldpad&SCE_PADLup) ))
		{
			XMEE_QuickShutdown(XM_ID[CS],XM_SONG_ID[CS]);	// Stops song, Free's Song,XMBuffer and XMPattern IOP RAM.
		}

// Perform full init of XM data (Sample data already in SPU RAM)
		else if ((paddata & SCE_PADLdown)&&(!(oldpad&SCE_PADLdown) ))
		{
			XM_IOPAddress[CS]=XMEE_QuickInit(XM_ID[CS],XM_SONG_ID[CS],XM_ADDR[CS],XM_SIZE[CS],XM_UseXMPanning);
		}

	    oldpad = paddata;
	}
	return(0);
}

/*****************************************************************************
TestTransfer
	Load XM files and sample files.
	Initialise everything
	Play song.
*****************************************************************************/
int TestTransfer(void)
{
int c,i;
int vb2;
int size;

// Load XM pattern data files, sample offset files and sample data files
	LoadFiles();

// Init XM system (Once off call..)
	XM_XMInit();

	XM_CreateIOPTransferBuffer();	// Creates a 64K buffer for sample data to be transfered from EE to IOP (to SPU)
	XM_SetSPUBaseAddress(0x05010);	// Set base address for transfer of sample data
	FlushIOPCommand(0);				// Process all XM_ calls on IOP (and wait for completion)
	Buffer64K=RPCReturnBuffer[4];	// Copy IOP allocated address for 64K buffer

// Create list of sample SPU RAM addresses, starting at SPU Base address.
// Note: These two calls can be either way around..
// Also note that the sample transfer/allocation does not use any IOP/SPU malloc type calls, so there is no
// need to close anything down when exiting a program..
	XMEE_CreateSampleOffsetData(XM_SAMPLEBANK_ID[0],(u_char *)SAMPLEHEADER_ADDR[0]);
	XMEE_TransferSampleData(SAMPLEBANK_ADDR[0],Buffer64K,SAMPLEBANK_SIZE[0]);		// Transfer the sample data to SPU via IOP

// And once again for the other song's sample files...
	XMEE_CreateSampleOffsetData(XM_SAMPLEBANK_ID[1],(u_char *)SAMPLEHEADER_ADDR[1]);
	XMEE_TransferSampleData(SAMPLEBANK_ADDR[1],Buffer64K,SAMPLEBANK_SIZE[1]);

// These two lines here will do the transfering to IOP, init xm data and song data..
	XM_IOPAddress[0]=XMEE_QuickInit(XM_ID[0],XM_SONG_ID[0],XM_ADDR[0],XM_SIZE[0],XM_UseXMPanning);
	XM_IOPAddress[1]=XMEE_QuickInit(XM_ID[1],XM_SONG_ID[1],XM_ADDR[1],XM_SIZE[1],XM_UseXMPanning);

// Free memory for the VB and VH files now. (samples now in SPU RAM)
// Note: We could also close the XM file too, but the demo code still needs this for reinitialising files...
	free((void*)SAMPLEBANK_ADDR[0]);
	free((void*)SAMPLEHEADER_ADDR[0]);
	free((void*)SAMPLEBANK_ADDR[1]);
	free((void*)SAMPLEHEADER_ADDR[1]);

// Free the 64K IOP Transfer buffer
	XM_FreeIOPTransferBuffer();

// Now Play it the xm files
	XM_XMPlay(XM_SAMPLEBANK_ID[0],XM_ID[0],XM_SONG_ID[0],0,XM_Loop,-1,0);
	FlushIOPCommand(SIF_RPCM_NOWAIT); 	// Transfer IOP Buffer
}


/*****************************************************************************
LoadFiles
	Load XM and Sample files into memory
*****************************************************************************/
int LoadFiles(void)
{
int fh,size;
u_int *filename;
u_int *paddress;
u_int *psize;
int a;

	filename=(u_int *)SAMPLEBANK_FILENAME;
	paddress=&SAMPLEBANK_ADDR[0];
	psize=&SAMPLEBANK_SIZE[0];
	LoadSingleFile(filename,paddress,psize);
	filename=(u_int *)SAMPLEHEADER_FILENAME;
	paddress=&SAMPLEHEADER_ADDR[0];
	psize=&SAMPLEHEADER_SIZE[0];
	LoadSingleFile(filename,paddress,psize);
	filename=(u_int*)XM_FILENAME;
	paddress=&XM_ADDR[0];
	psize=&XM_SIZE[0];
	LoadSingleFile(filename,paddress,psize);

	filename=(u_int *)SAMPLEBANK_FILENAME2;
	paddress=&SAMPLEBANK_ADDR[1];
	psize=&SAMPLEBANK_SIZE[1];
	LoadSingleFile(filename,paddress,psize);
	filename=(u_int *)SAMPLEHEADER_FILENAME2;
	paddress=&SAMPLEHEADER_ADDR[1];
	psize=&SAMPLEHEADER_SIZE[1];
	LoadSingleFile(filename,paddress,psize);
	filename=(u_int*)XM_FILENAME2;
	paddress=&XM_ADDR[1];
	psize=&XM_SIZE[1];
	LoadSingleFile(filename,paddress,psize);

}

/*****************************************************************************
LoadSingleFile
	Load a file
*****************************************************************************/
int LoadSingleFile(u_char *filename,u_int *address,u_int *sizep)
{
int fh,a,size;
u_int addr;

    if ((fh = sceOpen (filename, SCE_RDONLY)) < 0)
    {
		printf("file open failed. %s \n", filename);
		return -1;
    }

	size=sceLseek(fh,0,SCE_SEEK_END);
	*sizep=size;
    addr=(unsigned short int*)memalign(64,size);	// Buffer for loaded data
	*address=addr;

	sceLseek(fh,0,SCE_SEEK_SET);

	a=sceRead(fh,(unsigned char *) addr,size);
	if (a!=size)
    {
		printf("FileLoad fail\n");
		close (fh);
		return -1;
    }
	sceClose (fh);
	return(size);
}

