/***************************************************************************** 
XMPLAY.C
	XM (Extended MOD) Player code.
	PS2 Version (Based on the PS1 version.)
	Written by Jason Page @ SCEE
	Last Update 17th November 2000

****************************************************************************/
#include "/usr/local/sce/iop/install/include/OLD/timerold.h"
#include <kernel.h>
#include <sys/types.h>
#include <sys/file.h>
#include <stdio.h>
#include <string.h>
#include <sif.h>
#include <sifcmd.h>
#include <sifrpc.h>
#include <libsd.h>
#include <libcdvd.h>

#include "xmplay.h"		// structures, defines...
#include "xmdata.h"		// pitch tables
#include "xmcalls.h"	// EE->IOP call instructions

static int makeMyThread( void );
int UpdateXM(int status);
static void* ProcessEECommand (unsigned int command, void *data_, int size);

typedef struct TimerCtx
{
	int threadID;
	int timerID;
} TimerCtx;

int TimerFunc( TimerCtx* timer );
void startTimer( TimerCtx* timer );
void setTimer( TimerCtx* timer );
TimerCtx timer;

int VCount=0;
int gThid = 0;		// Thread ID
int gSem = 0;

int EE_SongFeedbackID=0;
int VAG_Addr=0x05010;
unsigned long xm_l_vag_spu_addr[8][128];  /* Address in memory for first sound file */


/**** XM SPECIFIC ****/ 
int CoreNum;
int CoreSD;
int UserCoreNum;
int UserCoreSD;

int JP_Do_Nothing; 
 
#define BPMLimit 256

char MonoMode=XM_STEREO;
int DVL;
int DVR;

int XM_FrameSpeed;
XMCHANNEL *XMCU;
XMCHANNEL *XMC;
XMHEADER *mh;					/* Pointer used by internal routines */
XMHEADER *mhu;					/* Pointer used by User routines */
XMSONG	*ms;					/* Pointer used by internal routines */
XMSONG	*mu;					/* Pointer used by User routines */


int		XM_HA=0;				// NEXT HEADER Address;
u_char	*XM_SngAddress[32];
u_char	*XM_HeaderAddress[8];

#define getWord(mpp) ((u_short)((*(u_char *)(mpp)) + ((*((u_char *)(mpp) + 1)) << 8)))
#define getSWord(mpp) ((short)((*(u_char *)(mpp)) + ((*((u_char *)(mpp) + 1)) << 8)))

int CurrentCh;


short XMSongIDs[32]=
					 {-1,-1,-1,-1,-1,-1,-1,-1,
					  -1,-1,-1,-1,-1,-1,-1,-1,	
					  -1,-1,-1,-1,-1,-1,-1,-1,	
					  -1,-1,-1,-1,-1,-1,-1,-1};


/*****************************************************************************
Interpolate
	Requires:
				p	CurrentPos
				p1	Start Pos
				p2 End Pos
				v1 Original Value
				v2 Destination value
*****************************************************************************/
short Interpolate(short p,short p1,short p2,short v1,short v2)
{
short dp,dv,di;

	if(p1==p2) return v1;

	dv=v2-v1;
	dp=p2-p1;
	di=p-p1;

	return v1 + ((long)(di*dv) / dp);
}

/*****************************************************************************
GetLogPeriod
	Returns AMIGA period, calculated from note and fine.
*****************************************************************************/
u_short GetLogPeriod(u_char note,u_short fine)
{
u_char n,o;
u_short p1,p2,i;
short j;
int f;

	n=note%12;
	o=(note/12);

	i=fine>>4;
	i+=(n<<3);                     /* n*8 + fine/16 */

	p1=logtab[i];
	p2=logtab[i+1];

	j=(Interpolate(fine/16,0,15,p1,p2)>>o);

	if (j==0)
	{
		f=1;
	}

   return(j);

}

/*****************************************************************************
JPGetPeriod
	Returns linear period calculated from note and fine.
*****************************************************************************/
u_short JPGetPeriod(u_char note,short fine)
{
	return((10L*12*16*4)-((u_short)note*16*4)-(fine/2)+64);
}

/*****************************************************************************
GetPeriod
	Returns period using either linear or Amiga(log) frequency
*****************************************************************************/
u_short GetPeriod(u_char note,u_short c2spd)
{
int a;
	a=mh->flags;
	if (ms->NotAmiga==1)
		return (JPGetPeriod(note,c2spd));		// Linear Freq table
	else
		return (GetLogPeriod(note,c2spd));		// Amiga Freq table
}

/***************************************************************************** 
XM_Exit
	Stop all tunes. 
	Clear SPU of all samples used. 
*****************************************************************************/ 
void XM_Exit(void) 
{ 
int i; 
	for (i=0;i<32;i++) 
	{ 
		XM_Quit(i); 
	} 
} 

/***************************************************************************** 
XM_PlayStop
	Stops a currently playing song.
	Silences SPU.
	Called by XM_Quit
*****************************************************************************/
void XM_PlayStop(int Song_ID)
{
	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]!=0)
	{
		mu=(XMSONG*)(XM_SngAddress[Song_ID]);
		if((mu->XMPlay!=XM_PLAYING)&&(mu->XMPlay!=XM_PAUSED))
			return;
		mu->XMPlay=XM_STOPPED;
		SilenceXM(Song_ID);
	}
	else
	{
		printf("aaaaa\n");
	}
}

/*****************************************************************************
InitXMData
	Initialise an XM file.
	Requires:
				Address of XM File
				XM ID (up to user 0-23)
				S3MPan switch (XM_UseS3MPanning or XM_UseXMPanning)
*****************************************************************************/
int InitXMData(u_char *mpp,int XM_ID,int S3MPan)
{
int t;
int a;
int c;

u_long b;
u_short b2;

	mhu=(XMHEADER*)(XM_HeaderAddress[XM_ID]);

	mhu->S3MPanning =S3MPan;

	mhu->version		=getWord(mpp+MD_version);
	
	mhu->songlength	=getWord(mpp+MD_snglen);
	mhu->restart		=getWord(mpp+MD_restart);
	mhu->XMChannels	=getWord(mpp+MD_numchan);
	mhu->numpat		=getWord(mpp+MD_numpat);
	mhu->numins		=getWord(mpp+MD_numins);
	mhu->flags		=getWord(mpp+MD_flags);
	mhu->tempo		=getWord(mpp+MD_tempo);
	mhu->bpm		=getWord(mpp+MD_bpm);
	mhu->XMPSXChannels=mhu->XMChannels;
	if (mhu->XMPSXChannels>32)
		mhu->XMPSXChannels=32;


	for (t=0;t<mhu->songlength;t++)
	{
		mhu->jorders[t]=((u_char)*(mpp+MD_orders+t));
	}

	a=MD_patstrt;
	c=mhu->numpat;

	for(t=0;t<c;t++)
	{
		mhu->JAP_PAT_ADDR[t]=(u_long*)(mpp+a);				/* Store Pattern Addr's */
		b=GetLong(mpp+a);
		mhu->JAP_PAT_ADDR2[t]=(u_long*)(mpp+a+b);			/* Store Pattern Addr's */
		b2=getWord(mpp+a+7);
		b+=b2;
		a+=b;
	}

	c=mhu->numins;
	for (t=0;t<c;t++)
	{
		b=GetLong(mpp+a);
		b2=getWord(mpp+a+27);
		if (b2!=0)
		{
			mhu->JAP_SampAddr[t]=(u_long*)(mpp+29+a);
			mhu->JAP_SampHdrAddr[t]=(u_long*)(mpp+b+a);
			b+=(40*b2);
		}
		else
		{
			mhu->JAP_SampAddr[t]=0xcdcdcdcd;	//(u_long*)(mpp+29+a);
			mhu->JAP_SampHdrAddr[t]=0x01234567;	//(u_long*)(mpp+b+a);
		}
		mhu->JAP_InstrumentOffset[t]=(u_long*)(mpp+a);		/* Store Instrument Addr's*/
		a+=b;
	}

	return(mhu->XMPSXChannels);
}

/*****************************************************************************
GetWord
	Returns a word from given address (needed cos data may be on odd boundries)
*****************************************************************************/
u_short GetWord(u_char *mpp)
{
u_char a;
u_char b;
u_short d;

	a=*(mpp);
	b=*(mpp+1);
	d=a+(b<<8);
	return(d);
}

/*****************************************************************************
GetSWord
	Returns an s-word from given address
*****************************************************************************/
short GetSWord(u_char *mpp)
{
u_char a;
u_char b;
short f;

	a=*(mpp);
	b=*(mpp+1);

	f=a+(b<<8);
	return(f);
}

/*****************************************************************************
GetLong
	Returns a long from given address
*****************************************************************************/

u_long GetLong(u_char *mpp)
{
u_char a;
u_char b;
u_char c;
u_char d;
u_long e;

	a=*(mpp);
	b=*(mpp+1);
	c=*(mpp+2);
	d=*(mpp+3);

	e=a+(b<<8)+(c<<16)+(d<<24);

	return(e);
}

/*****************************************************************************
XM_OnceOffInit
	Initialises system. Sets up for 50 or 60 fps playback
*****************************************************************************/
void XM_OnceOffInit(void)
{
int i;
	for (i=0;i<32;i++)
	{
		XM_SngAddress[i]=0;
	}
	CoreNum=-1;
	CoreSD=-1;
	UserCoreSD=-1;

	JP_Do_Nothing=0;				/* Allow XM_Update to process */
}

/*****************************************************************************
XM_Init
	Initialise a song or SFX.
	Requires:
			VAB_ID (0-7)		ID of VAB data (returned from PSX_Init)
			XM_ID (0-23)		Value corresponding to InitXMData call.
			FirstCh				First SPU Channel to play back on.
			Loop					XM_Loop or XM_NoLoop
			PlayMask				Bitpattern for which channels to process
			SFXNum				Pattern number to play, if XM_SFX
*****************************************************************************/
int  XM_Init(int VabID,int XM_ID,int SongID, int FirstCh,
				 int Loop,int PlayMask,int SFXNum)
{
int i;
int pmsk;
int Chn;
int SngID;


	if (SongID>31)
		return(-1);				// Not enough malloc'd song structures..
	XM_Quit(SongID);
	XMSongIDs[SongID]=0;
	SngID=SongID;

	if (XM_SngAddress[SngID]==0)
		return(-2);

	JP_Do_Nothing=1;
	mu=(XMSONG*)(XM_SngAddress[SngID]);
	mhu=(XMHEADER*)(XM_HeaderAddress[XM_ID]);

	mu->XMPlay=XM_STOPPED;

	mu->VabID=VabID;
	mu->SongLoop=Loop;
	mu->PlayMask=PlayMask;
	mu->SFXNum=SFXNum;

	mu->HeaderNum=XM_ID;
	mu->JBPM=0;
	mu->PCounter=3;
	mu->reppos=mhu->restart;
	mu->repcnt=0;
	mu->CurPos=0;
	mu->SongPos=0;
	mu->SongPos=SFXNum;
	mu->CurrentStart=SFXNum;
	mu->SongSpeed=mhu->tempo;
	mu->vbtick=mu->SongSpeed;
	mu->SongBPM=mhu->bpm;

	mu->NotAmiga=mhu->flags;
	mu->patdly=0;
	mu->patdly2=0;
	mu->SongVolume=128;
	mu->MasterVolume=128;
	mu->FirstCh=FirstCh;
	mu->PatternPos=0;
	mu->posjmp=1;			/* <- make sure the player fetches the first note */
	mu->patbrk=0;
	mu->PlayNext=-1;
	mu->UserPan=0;		/* Default Middle */
	mu->MaxChans=0;
	Chn=FirstCh;
	for(i=0;i<mhu->XMPSXChannels;i++)
	{
		pmsk=mu->PlayMask;
		pmsk&=(1<<i);
		if (pmsk!=0)
		{
			XMCU=&mu->XM_Chnl[i];		/* Build list of SPU Channels to use */
			XMCU->SPUChannel=Chn;		/* Depending on bitmask */
			mu->MaxChans++;
			if (Chn<48)
				InitSPUChannel(Chn);
			Chn++;
		}
	}

	for(i=0;i<mhu->XMChannels;i++)
	{
		mu->XM_Chnl[i].UserVol=0;
		mu->XM_Chnl[i].kick=0;
		mu->XM_Chnl[i].kick2=0;
		mu->XM_Chnl[i].LVol=0;
		mu->XM_Chnl[i].RVol=0;
		mu->XM_Chnl[i].volume=0+64;
		mu->XM_Chnl[i].tmpvolume=0+64;
		mu->XM_Chnl[i].retrig=0;
		mu->XM_Chnl[i].wavecontrol=0;
		mu->XM_Chnl[i].glissando=0;
		mu->XM_Chnl[i].panning=128;
		mu->XM_Chnl[i].note=0;
		mu->XM_Chnl[i].vol=0;
		mu->XM_Chnl[i].eff=0;
		mu->XM_Chnl[i].dat=0;
		mu->XM_Chnl[i].ins=0;
		mu->XM_Chnl[i].ChDead=1;
		mu->XM_Chnl[i].nothing=1;
		mu->XM_Chnl[i].OldLVol=-1;
		mu->XM_Chnl[i].OldRVol=-1;
		mu->XM_Chnl[i].OldPeriod=0;
		mu->XM_Chnl[i].sample=254;
		mu->XM_Chnl[i].OldSample=-1;
	}
  	mu->XMPlay=XM_PLAYING;

	JP_Do_Nothing=0;
	return(SngID);
}

/*****************************************************************************
JPlayNote
	Decode Pattern data.
	Setup note/instrument/volume/effect
	Return amount of bytes read.
*****************************************************************************/
int JPlayNote(u_char *j,int pmsk)
{
u_char b;
int ret=1;
u_char note=0;

	b=*j;

	XMC->eff=0;
	XMC->vol=0;
	XMC->dat=0;

	if (b==0x80)
	{
		XMC->nothing=1;
		return(ret);
	}

	j++;

	if ((b&0x80)!=0)
	{
		if ((b&1)!=0)
		{
			if (pmsk!=0)
			{
				XMC->nothing=0;
				note=*(j);
				SetNote(note-1);
			}
			j++;
			ret++;
		}

		if ((b&2)!=0)
		{
			if (pmsk!=0)
			{
				XMC->ins=*(j);
				SetInstr(XMC->ins-1);
			}
			j++;
			ret++;
		}
		if (((b&3)!=0)&&(note!=97)&&(pmsk!=0))
			SetPer();			/* Set Pitch if new instrument/note */

		if ((b&4)!=0)
		{
			XMC->nothing=0;
			XMC->vol=*(j);
			j++;
			ret++;
		}
		
		if ((b&8)!=0)
		{
			XMC->nothing=0;
			XMC->eff=*(j);
			j++;
			ret++;
		}
		if ((b&16)!=0)
		{
			XMC->dat=*(j);
			ret++;
		}
		
	}
	else
	{
		XMC->nothing=0;
		note=b;

		if (pmsk!=0)
		{
			SetNote(note-1);
			XMC->ins=*(j);
			SetInstr(XMC->ins-1);
			if (note!=97)
				SetPer();
			XMC->vol=*(j+1);
			XMC->eff=*(j+2);
			XMC->dat=*(j+3);
		}
		ret=5;
	}																				 
	return(ret);
}

/*****************************************************************************
JPlayEffects
	Checks if any effects to be played.
	Calls relevent routines if thay are.
*****************************************************************************/
void JPlayEffects(void)
{
u_char note=0;
u_char ins=0;
u_char vol=0;
u_char eff=0;
u_char dat=0;

	note=XMC->note;
	vol=XMC->vol;
	eff=XMC->eff;
	dat=XMC->dat;
	ins=XMC->ins;

	XMC->ownper=0;
	XMC->ownvol=0;

	switch(vol>>4)
	{ 
		case 0x6:
			if(vol&0xf)
				SPE(0xa,vol&0xf);				/* Volume slide down */
			break;
	
		case 0x7:
			if(vol&0xf)
				SPE(0xa,vol<<4);				/* Volume slide up */
			break;
	
		case 0x8:
			SPE (0xe,0xb0 | (vol&0xf));	/* Fine volume slide down */
			break;

		case 0x9:                     
			SPE (0xe,0xa0 | (vol&0xf));	/* Fine volume slide up */
			break;

		case 0xa:
			SPE (0x4,vol<<4);					/* Set Vibrato speed */
			break;

		case 0xb:
			SPE (0x4,vol&0xf);				/* Vibrato */
			break;

		case 0xc:
			XMC->panning=(vol&0xf)<<4;		/* Set Panning */
			break;

		case 0xd:
			if(vol&0xf)
				DoXMPanSlide(vol&0xf);		/* Pan slide left */
			break;

		case 0xe:
			if(vol&0xf)
				DoXMPanSlide(vol<<4);		/* Pan slide right */
			break;

		case 0xf:
			SPE (0x3,vol<<4);					/* Portamento */
			break;

		default:
			if(vol>=0x10 && vol<=0x50)
			{
				SPE (0xc,vol-0x10);			/* Set volume */
			}
	}
		switch(eff)
			{
			case 'G'-55:                    /* G - set global volume */
				ms->SongVolume=dat*2;
				break;

			case 'H'-55:                    /* H - global volume slide */
				break;

			case 'K'-55:                    /* K - keyoff */
					SetNote(96);
				break;

			case 'L'-55:                    /* L - set envelope position */
				break;

			case 'P'-55:                    /* P - panning slide */
				XMC->panning=dat;
				break;

			case 'R'-55:                    /* R - multi retrig note */
				DoS3MRetrig(dat);
				break;

			case 'T'-55:
				SPE (0x6,dat);
				break;

			case 'X'-55:
				break;

			default:
				if(eff<=0xf)
					SPE (eff,dat);				/* Do most of the effects */
				break;
		}

	if(!XMC->ownper)
	{
		XMC->period=XMC->tmpperiod;
	}

	if(!XMC->ownvol)
	{
		XMC->volume=XMC->tmpvolume;
	}
}

/*****************************************************************************
SPE
	This is the bit that does most of the updating of effects
*****************************************************************************/
void SPE (u_char eff,u_char dat)
{
int hi;
int lo;

	switch(eff)
	{
		case XMEF_APPREGIO:		/* 0 */
			Arpeggio(dat);
			break;

		case XMEF_PORTUP:			/* 1 */
			if(dat!=0)
				XMC->slidespeed=((u_short)(dat))<<2;
			if(ms->vbtick)
			{
				XMC->tmpperiod-=XMC->slidespeed;

			}
			break;

		case XMEF_PORTDOWN:		/* 2 */
			if(dat!=0)
				XMC->slidespeed=(u_short)(dat)<<2;
			if(ms->vbtick)
			{
				XMC->tmpperiod+=XMC->slidespeed;
			}
			break;

		case XMEF_TONEPORT:		/* 3 */
			XMC->kick=0;
			if(dat!=0)
			{
				XMC->portspeed=dat;
				XMC->portspeed<<=2;
			}
			DoToneSlide();
			XMC->ownper=1;
			break;

		case XMEF_VIBRATO:		/* 4 */
			if(dat&0x0f) XMC->vibdepth=dat&0x0f;
			if(dat&0xf0) XMC->vibspd=(dat&0xf0)>>2;
			DoVibrato();
			XMC->ownper=1;
			break;

		case XMEF_PORT_VOLSLD:	/* 5 */
			XMC->kick=0;
			DoToneSlide();
			if (dat==0)
				dat=XMC->oldvslide;
			XMC->oldvslide=dat;
			DoVolSlide(dat);
			XMC->ownper=1;
			break;

		case XMEF_VIB_VOLSLD:	/* 6 */
			DoVibrato();
			if (dat==0)
				dat=XMC->oldvslide;
			XMC->oldvslide=dat;
			DoVolSlide(dat);
			XMC->ownper=1;
			break;

		case XMEF_TREMELO:		/* 7 */
			if(dat&0x0f) XMC->trmdepth=dat&0xf;
			if(dat&0xf0) XMC->trmspd=(dat&0xf0)>>2;
			DoTremolo();
			XMC->ownvol=1;
			break;

		case XMEF_PANPOS:			/* 8 */
			XMC->panning=dat;
			break;

		case XMEF_SAMPOFFSET:	/* 9 */
			if(ms->vbtick)
				break;
			if(dat==0)
				break;

			XMC->SOffset=dat;
			XMC->SOffset*=256;
			XMC->SOffset*=2;
			XMC->SOffset/=7;
			XMC->SOffset*=2;
			XMC->SOffset&=(~31);
			break;

		case XMEF_VOLSLD:			/* A */
			if (dat==0)
				dat=XMC->oldvslide;
			XMC->oldvslide=dat;
			DoVolSlide(dat);
			break;

		case XMEF_POSJMP:			/* B */
			if(ms->patdly2)
				break;
			ms->patbrk=0;
			ms->SongPos=dat-1;
			ms->posjmp=2;
			break;

		case XMEF_VOLUME:			/* C */
			if(ms->vbtick)
				break;
			if(dat>64)
				dat=64;
			XMC->tmpvolume=dat+64;
			break;

		case XMEF_PATBREAK:		/* D */
			if(ms->patdly2)
				break;
			hi=(dat&0xf0)>>4;
			lo=(dat&0xf);
			ms->patbrk=((hi*10)+lo)+1;

			if(ms->patbrk>(ms->numrow+1))
				ms->patbrk=ms->numrow+1;
			ms->posjmp=2;
			break;

		case XMEF_E:				/* E */
			DoEEffects(dat);
			break;

		case XMEF_SETSPEED:		/* F */
			if(ms->vbtick || ms->patdly2)
				break;
			if(dat<0x20)
			{
				ms->SongSpeed=dat;	// Speed
				ms->vbtick=0;
			}
			else
			{
				ms->SongBPM=dat;		// BPM
				ms->vbtick=0;
			}
			break;

		default:
			break;
	}
}

/*****************************************************************************
DoEEffects
	Process effects if 0xe command used.
*****************************************************************************/
void DoEEffects(u_char dat)
{
u_char nib;

	nib=dat&0xf;

	switch(dat>>4)
	{
		case 0x0:							/* 0 filter toggle */
				break;

		case XMEF_E_FINESLD_UP:			/* 1 Fineslide up */
				if(!ms->vbtick)
				{
					if (nib==0)
						nib=XMC->oldfslide;
					XMC->oldfslide=nib;
					XMC->tmpperiod+=(nib<<2);
				}
				break;

		case XMEF_E_FINESLD_DOWN:		/* 2 Fineslide down */
				if(!ms->vbtick)
				{
					if (nib==0)
						nib=XMC->oldfslide;
					XMC->oldfslide=nib;
					XMC->tmpperiod-=(nib<<2);
				}

				break;

		case XMEF_E_GLISSANDO:			/* 3 Glissando ctrl */
				XMC->glissando=nib;
				break;

		case XMEF_E_VIB_WAVE:			/* 4 Set vibrato waveform */
				XMC->wavecontrol&=0xf0;
				XMC->wavecontrol|=nib;
				break;

		case XMEF_E_FINETUNE:			/* 5 Set finetune */
				break;

		case XMEF_E_PATLOOP:				/* 6 Set patternloop */

				if(ms->vbtick)
					break;
				if(nib)
				{
					if(ms->repcnt>0)
						ms->repcnt--;		/* already looping, decrease counter */
					else
						ms->repcnt=nib;    /* not yet looping, so set repcnt */

					if(ms->repcnt)        /* jump to reppos if repcnt>0 */
						ms->PatternPos=ms->reppos;
				}
				else
				{
					ms->reppos=ms->PatternPos-1;	/* set reppos */
				}
				break;

		case XMEF_E_TREM_WAVE:			/* 7 Set tremolo waveform */
				XMC->wavecontrol&=0x0f;
				XMC->wavecontrol|=nib<<4;
				break;

		case XMEF_E_NOTUSED:				/* 8 Not used */
				XMC->panning=nib<<4;
				break;

		case XMEF_E_RETRIG:				/* 9 Retrig note */

				/* only retrigger if
				   data nibble > 0 */

				if(nib>0)
				{
					if(XMC->retrig==0)
					{
						XMC->kick=1;
						XMC->kick2=1;
						XMC->retrig=nib;
					}
					XMC->retrig--; /* countdown */
				}
				break;

		case XMEF_E_FINEVOL_UP:			/* A - Fine volume slide up */
				if(ms->vbtick)
					break;

				if (nib==0)
					nib=XMC->oldfvslide;
				XMC->oldfvslide=nib;
				XMC->tmpvolume+=nib;
				if(XMC->tmpvolume>128) XMC->tmpvolume=128;
				break;

		case XMEF_E_FINEVOL_DOWN:		/* B - Fine volume slide down */
				if(ms->vbtick)
					break;

				if (nib==0)
					nib=XMC->oldfvslide;
				XMC->oldfvslide=nib;
				XMC->tmpvolume-=nib;
				if(XMC->tmpvolume<64) XMC->tmpvolume=64;
				break;

		case XMEF_E_CUTNOTE:				/* C - Cut note */

				if(ms->vbtick>=nib)
				{
					XMC->tmpvolume=64;
				}
				break;

		case XMEF_E_NOTEDELAY:			/* D - Note delay */

				if(ms->vbtick==nib)
				{
					XMC->kick=1;
					XMC->kick2=1;
				}
				else
					XMC->kick=0;
				break;

		case XMEF_E_PATDELAY:			/* E - Pattern delay */
				if(ms->vbtick)
					break;
				if(!ms->patdly2)
					ms->patdly=nib+1;		/* only once (when vbtick=0) */
				break;

		case 0xf:       /* invert loop, not supported */
				break;
	}
}

/*****************************************************************************
SetNote
	Set note if a new note is to processed
*****************************************************************************/
void SetNote(u_char note)
{
u_char* j;
u_char vf;

	if (note==96)
	{ 
		XMC->keyon=0;
		if (XMC->sample==254)
		{
			XMC->tmpvolume=64;			/* Force 0 vol if no envelope used */
		}
		else
		{
			j=mh->JAP_InstrumentOffset[XMC->sample];
			vf=*((u_char*)j+233);
			if(!(vf & EF_ON))
			{
				XMC->tmpvolume=64;			/* Force 0 vol if no envelope used */
			}
		}
	}
	else
	{
		XMC->note=note;
		XMC->kick=1;
		XMC->kick2=1;

		if(!(XMC->wavecontrol&0x80))	/* retrig tremolo and vibrato waves ? */
			XMC->trmpos=0;
		if(!(XMC->wavecontrol&0x08))
			XMC->vibpos=0;
	}
}

/*****************************************************************************
SetInstr
	Set instrument if new one to be processed
	Copy values from XM data (transpose, fine,volume...)
*****************************************************************************/

void SetInstr(u_char inst)
{
u_long *j;
u_char j2;
short ddd;
u_int j3;

	if (!XMC->kick)
		return;							// Instrument but no note...

	if (inst==255)
	{
		inst=XMC->sample;				// If instrument=0, use last instrument.
	}

	if(inst>=mh->numins) return;             // No Instrument available

	XMC->sample=inst;
  
	j=mh->JAP_SampHdrAddr[inst];
	if (j==0x01234567)
		return;

	j2=*((u_char*)j+12);
	XMC->tmpvolume=j2+64;
	XMC->volume=j2+64;
	XMC->vol=j2+0x10;
	XMC->retrig=0;

	XMC->LoopFlag=*((u_char*)j+14);

	if (mh->S3MPanning==XM_UseXMPanning)
		XMC->panning=*((u_char*)j+15);

	j3=*((u_char*)j+7);
	j3<<=8;
	j3+=*((u_char*)j+6);
	j3<<=8;
	j3+=*((u_char*)j+5);
	j3<<=8;
	j3+=*((u_char*)j+4);
	XMC->LoopStart=j3;
//	if ((j3!=0)&&(XMC->LoopFlag==1))
//		printf("ins %x,looppos =%x\n",inst,XMC->LoopStart);

	XMC->transpose=*((u_char*)j+16);
	ddd=(*((u_char*)j+13));
	ddd+=128;
	XMC->c2spd=ddd;
	XMC->c2spd&=255;
}

/*****************************************************************************
SetPer
	Init insturment envelopes, period...
	Called if either SetInst or SetNote are called
*****************************************************************************/
void SetPer(void)
{
u_long *j;
u_short period;
u_char a,b;
u_short tmp;

	if (CurrentCh>31)
		return;

	a=XMC->note;
	a+=XMC->transpose;

	period=GetPeriod(a,XMC->c2spd);

	XMC->wantedperiod=period;
	XMC->tmpperiod=period;
	XMC->OldPeriod=0;
	XMC->SOffset=0;

	if (XMC->OldSample!=XMC->sample)
	{
		XMC->OldSample=XMC->sample;
		j=mh->JAP_InstrumentOffset[XMC->sample];

		b=*((u_char*)j+240);	//39
		tmp=b<<8;
		b=*((u_char*)j+239);	//39
		tmp+=b;

		XMC->keyoffspd=tmp;

		XMC->envflg=*((u_char*)j+233);
		if(XMC->envflg & EF_ON)
		{
			XMC->envpts=*((u_char*)j+225);
			XMC->envsus=*((u_char*)j+227);
			XMC->envbeg=*((u_char*)j+228);
			XMC->envend=*((u_char*)j+229);
		}
		XMC->panenvflg=*((u_char*)j+234);
		if(XMC->panenvflg & EF_ON)
		{
			XMC->panenvpts=*((u_char*)j+226);
			XMC->panenvsus=*((u_char*)j+230);
			XMC->panenvbeg=*((u_char*)j+231);
			XMC->panenvend=*((u_char*)j+232);
		}
	}
	if(XMC->panenvflg & EF_ON)
	{
		XMC->panenvp=0;
		XMC->panenva=0;
		XMC->panenvb=1;
	}
	if(XMC->envflg & EF_ON)
	{
		XMC->envp=0;
		XMC->enva=0;
		XMC->envb=1;
	}
}

/*****************************************************************************
Arpeggio
	EFFECT - Modify pitch
*****************************************************************************/
void Arpeggio(u_char dat)
{
	u_char note;

	note=XMC->note;

	if(dat!=0)
	{
		switch(ms->vbtick%3)
		{
			case 1:
				note+=(dat>>4);
				break;
			case 2:
				note+=(dat&0xf);
				break;
		}
		XMC->period=GetPeriod(note+XMC->transpose,XMC->c2spd);
		XMC->ownper=1;
	}
}

/*****************************************************************************
DoVolSlide
	EFFECT - Either +/- 0-0xf to volume
*****************************************************************************/
void DoVolSlide(u_char dat)
{
	if(!ms->vbtick) return;				 /* do not update when vbtick==0 */

	XMC->tmpvolume+=dat>>4;           /* volume slide */
	if(XMC->tmpvolume>128) XMC->tmpvolume=128;
	XMC->tmpvolume-=dat&0xf;
	if(XMC->tmpvolume<64) XMC->tmpvolume=64;
}

/*****************************************************************************
DoXMPanSlide
	EFFECT - Pan slide
*****************************************************************************/
void DoXMPanSlide(u_char inf)
{
u_char lo,hi;
short pan;

	if(inf!=0) XMC->pansspd=inf;
	else inf=XMC->pansspd;

	if(!ms->vbtick) return;

	lo=inf&0xf;
	hi=inf>>4;

	/* slide right has absolute priority: */

	if(hi)
		lo=0;

	pan=XMC->panning;

	pan-=lo;
	pan+=hi;

	if(pan<0) pan=0;
	if(pan>255) pan=255;

	XMC->panning=pan;
}

/*****************************************************************************
DoS3MRetrig
	EFFECT - Pan slide
*****************************************************************************/
void DoS3MRetrig(u_char inf)
{
	u_char hi,lo;

	hi=inf>>4;
	lo=inf&0xf;

	if(lo)
	{
		XMC->s3mrtgslide=hi;
		XMC->s3mrtgspeed=lo;
	}

	if(hi)
	{
		XMC->s3mrtgslide=hi;
	}

	/* only retrigger if lo nibble > 0 */

	if(XMC->s3mrtgspeed>0)
	{
		if(XMC->retrig==0)
		{

			/* when retrig counter reaches 0,
			   reset counter and restart the sample */

			XMC->kick=1;
			XMC->kick=2;
			XMC->retrig=XMC->s3mrtgspeed;

			if(ms->vbtick)
			{                     /* don't slide on first retrig */
				switch(XMC->s3mrtgslide)
				{

					case 1:
					case 2:
					case 3:
					case 4:
					case 5:
						XMC->tmpvolume-=(1<<(XMC->s3mrtgslide-1));
						break;

					case 6:
						XMC->tmpvolume=(2*XMC->tmpvolume)/3;
						break;

					case 7:
						XMC->tmpvolume=XMC->tmpvolume>>1;
						if (XMC->tmpvolume<64)
							XMC->tmpvolume=64;
						break;

					case 9:
					case 0xa:
					case 0xb:
					case 0xc:
					case 0xd:
						XMC->tmpvolume+=(1<<(XMC->s3mrtgslide-9));
						break;

					case 0xe:
						XMC->tmpvolume=(3*XMC->tmpvolume)/2;
						break;

					case 0xf:
						XMC->tmpvolume=XMC->tmpvolume<<1;
						break;
				}
				if(XMC->tmpvolume<64) XMC->tmpvolume=64;
				if(XMC->tmpvolume>128) XMC->tmpvolume=128;
			}
		}
		XMC->retrig--; /* countdown */
	}
}

/*****************************************************************************
DoToneSlide
	EFFECT
*****************************************************************************/
void DoToneSlide(void)
{
int dist;

	/* We have to slide XMC->period towards XMC->wantedperiod, so
	   compute the difference between those two values */

	dist=XMC->period-XMC->wantedperiod;

	if(dist==0 || XMC->portspeed>abs(dist))/* if they are equal */
	{											      /* or if portamentospeed is too big */

		XMC->period=XMC->wantedperiod;      /* make tmpperiod equal tperiod */
	}
	else if(dist>0)
	{		                                 /* dist>0 ? */
		XMC->period-=XMC->portspeed;        /* then slide up */
	}
	else
		XMC->period+=XMC->portspeed;        /* dist<0 -> slide down */

	XMC->tmpperiod=XMC->period;				/* Store new period */
}

/*****************************************************************************
DoVibrato
	EFFECT
*****************************************************************************/
void DoVibrato(void)
{
u_char q;
u_short temp=0;

	q=(XMC->vibpos>>2)&0x1f;

	switch(XMC->wavecontrol&3)
	{

		case 0: /* sine */
			temp=VibratoTable[q];
			break;

		case 1: /* ramp down */
			q<<=3;
			if(XMC->vibpos>127)
				q=255-q;
			temp=q;
			break;

		case 2: /* square wave */
			temp=255;
			break;
	}

	temp*=XMC->vibdepth;
	temp>>=7;
	temp<<=2;

	if(XMC->vibpos<128)
		XMC->period=XMC->tmpperiod+temp;
	else
		XMC->period=XMC->tmpperiod-temp;

	if(ms->vbtick) XMC->vibpos+=XMC->vibspd;        /* do not update when vbtick==0 */
}

/*****************************************************************************
DoTremelo
	EFFECT
*****************************************************************************/
void DoTremolo(void)
{
u_char q;
u_short temp=0;

	q=(XMC->trmpos>>2)&0x1f;

	switch((XMC->wavecontrol>>4)&3)
	{

		case 0: /* sine */
			temp=VibratoTable[q];
			break;

		case 1: /* ramp down */
			q<<=3;
			if(XMC->trmpos>127) q=255-q;
			temp=q;
			break;

		case 2: /* square wave */
			temp=255;
			break;
	}

	temp*=XMC->trmdepth;
	temp>>=6;

	if(XMC->trmpos<128)
	{
		XMC->volume=XMC->tmpvolume+temp;
		if(XMC->volume>128) XMC->volume=128;
	}
	else
	{
		XMC->volume=XMC->tmpvolume-temp;
		if(XMC->volume<64) XMC->volume=64;
	}

	if(ms->vbtick) XMC->trmpos+=XMC->trmspd;        /* do not update when vbtick==0 */
}

/*****************************************************************************
DoPan
	Calculate panning using current envelope panning and instrument pan
*****************************************************************************/
short DoPan(short envpan,short pan)
{
	return(pan + (((envpan-128)*(128-abs(pan-128)))/128));
}

/*****************************************************************************
DoVol
	Calculate volume using

	a Fade volume
	b Envelope volume
	c Instrument volume
*****************************************************************************/
short DoVol(u_long a,short b,short c)
{
	a*=b;
	a*=c;
	a>>=23;
	return(a);
}

/*****************************************************************************
UpdateXMData
	This does everything
*****************************************************************************/
void UpdateXMData(void)
{
int SC;


	if(JP_Do_Nothing)			/* Failsafe switch - if called on interrupt */
		return;

	for (SC=0;SC<32;SC++)
	{
		if (XM_SngAddress[SC]!=0)
		{
			ms=(XMSONG*)(XM_SngAddress[SC]);
			UpdateWithTimer(SC);
		}
	}
}

/*****************************************************************************
UpdateWithTimer
	Update BPM counter of song. Update everything else if its time..
*****************************************************************************/
void UpdateWithTimer(int SC)
{
	ms=(XMSONG*)(XM_SngAddress[SC]);
	if (ms->XMPlay!=XM_PLAYING)
		return;

	mh=(XMHEADER*)(XM_HeaderAddress[ms->HeaderNum]);
	ms->JBPM+=ms->SongBPM;
	if (ms->JBPM<BPMLimit)	// Time to process data ?
		return;

	ms->JBPM-=BPMLimit;		/* YES! Update song/hardware */
	XM_DoFullUpdate(SC);
}

/*****************************************************************************
XM_DoFullUpdate
	Update effects,hit hardware, read pattern data
*****************************************************************************/
void XM_DoFullUpdate(int SC)
{
	ms=(XMSONG*)(XM_SngAddress[SC]);
	if (ms->XMPlay==XM_PLAYING)
	{
		mh=(XMHEADER*)(XM_HeaderAddress[ms->HeaderNum]);

		UpdateEffs();			/* Do this if not done before in spare frame*/
		ApplyEffs();
		UpdateHardware();			/* Update SPU */
		if (ms->vbtick==1)		/* Check for zero volume,keyed off channels*/
			CurrentKeyStat();		/* BUT not on first tick - wait for keyons */
		UpdatePatternData(SC);
	}
}

/*****************************************************************************
UpdatePatternData
	Read new pattern data. Loop song / jump to a new pattern if necessary.
*****************************************************************************/
void UpdatePatternData(int SC)
{
int SP,t,pmsk;
int FindOffset;
u_char *pataddr;
u_char patdat;

	if(++ms->vbtick>=ms->SongSpeed)
	{
		ms->PatternPos++;		/* Time to read more pattern data */
		ms->vbtick=0;

		if(ms->patdly)
		{
			ms->patdly2=ms->patdly;
			ms->patdly=0;
		}

		if(ms->patdly2)
		{
			/* patterndelay active */
			if(--ms->patdly2)
			{
				ms->PatternPos--;    /* so turn back PatternPos by 1 */
			}
		}

/*
Do we have to get a new patternpointer ?
(when PatternPos reaches, say 64 or when a patternbreak is active)
*/

		FindOffset=0;

		if(ms->PatternPos == ms->numrow )
			ms->posjmp=2;

		if(ms->posjmp)
		{
			if (ms->patbrk!=0)
				ms->PatternPos=ms->patbrk-1;
			else
				ms->PatternPos=0;

			ms->CurPos=ms->PatternPos;
			if (ms->CurPos!=0)
				FindOffset=1;

			if (ms->PlayNext!=-1)
			{
				ms->SongPos=ms->PlayNext;
				ms->CurrentStart=ms->SongPos;
				ms->PlayNext=-1;
			}
			else
			{
				ms->SongPos+=(ms->posjmp-1);
				t=ms->SongPos;
			}

			ms->patbrk=ms->posjmp=0;
			if(ms->SongPos>=mh->songlength)		/* End of song data ? */
			{
				if(!ms->SongLoop)
				{
					ms->XMPlay=XM_STOPPED;			/* Once off tune */
					return;
				}
				ms->SongPos=ms->reppos;				/* Loop to loop point */
			}
			if(ms->SongPos<0)
				ms->SongPos=mh->songlength-1;
			SP=mh->jorders[ms->SongPos];		/* Pattern from song data*/
			ms->CurrentPattern=SP;
			ms->PatAdr=mh->JAP_PAT_ADDR[SP];		/* Addr of pat header */
			ms->PatAdr2=mh->JAP_PAT_ADDR2[SP];	/* Addr of pat data */
			ms->numrow=getWord((u_char*)ms->PatAdr+5);	/* Number of rows */
			ms->PatSize=getWord((u_char*)ms->PatAdr+7);	/* pattern size */
		}
		if(!ms->patdly2)
		{
			if (ms->PatSize>0)
			{
				if (mh->version==0xddba)				/*** SUPER COMPRESSED ***/
				{
					if (FindOffset!=0)
						ms->CurPos=PACKEDCalcPlayPos(ms->CurPos);
					SP=ms->CurPos;
					for (t=0;t<mh->XMPSXChannels;t++)
					{
						XMC=&ms->XM_Chnl[t];		/* Decode pattern data */
						XMC->eff=0;
						XMC->vol=0;
						XMC->dat=0;
					}

					while(1)
					{

						pataddr=((u_char*)ms->PatAdr2+SP);
						patdat=*pataddr++;				/* Channel or 0xff for end */
						SP++;
						if (patdat==0xff)
							break;

						CurrentCh=patdat;
						XMC=&ms->XM_Chnl[patdat];		/* Decode pattern data */
						pmsk=ms->PlayMask;
						pmsk&=(1<<patdat);
						SP+=JPlayNote((u_char*)ms->PatAdr2+SP,pmsk);
					}
					ms->CurPos=SP;

				}
				else
				{


					if (FindOffset!=0)					/**** ORIGINAL XM FILE ****/
						ms->CurPos=CalcPlayPos(ms->CurPos);
					SP=ms->CurPos;
					for (t=0;t<mh->XMChannels;t++)
					{
						CurrentCh=t;
						XMC=&ms->XM_Chnl[t];				/* Decode pattern data */
						pmsk=ms->PlayMask;
						pmsk&=(1<<t);

						SP+=JPlayNote((u_char*)ms->PatAdr2+SP,pmsk);
					}
					ms->CurPos=SP;
				}
			}
		}
	}
}

/*****************************************************************************
CalcPlayPos
	Calculate position through data if playing back from, say, halfway through
	a pattern
*****************************************************************************/
int CalcPlayPos(int StartPos)
{
int SP=0;
int t,i;

	for (i=0;i<StartPos;i++)
	{
		for (t=0;t<mh->XMChannels;t++)
		{
			CurrentCh=t;
			XMC=&ms->XM_Chnl[t];				/* Decode pattern data */
			SP+=JCalcPat((u_char*)ms->PatAdr2+SP);
		}
	}
	return(SP);
}

/*****************************************************************************
PACKEDCalcPlayPos
	As above but for PACKED format of data
*****************************************************************************/
int PACKEDCalcPlayPos(int StartPos)
{
int SP=0;
int i;
u_char *pataddr;
u_char patdat;

	for (i=0;i<StartPos;i++)
	{
		while(1)
		{
			pataddr=((u_char*)ms->PatAdr2+SP);
			patdat=*pataddr++;				/* Channel or 0xff for end */
			SP++;
			if (patdat==0xff)
				break;

			CurrentCh=patdat;
			XMC=&ms->XM_Chnl[patdat];		/* Decode pattern data */
			SP+=JCalcPat((u_char*)ms->PatAdr2+SP);
		}
	}
	return(SP);
}

/*****************************************************************************
JCalcPat
	Return the amount of bytes to skip
*****************************************************************************/
int JCalcPat(u_char *j)
{
u_char b;
int ret=1;

	b=*j;

	if (b==0x80)
		return(ret);
	j++;

	if ((b&0x80)!=0)
	{
		if ((b&1)!=0)
		{
			j++;
			ret++;
		}

		if ((b&2)!=0)
		{
			j++;
			ret++;
		}

		if ((b&4)!=0)
		{
			j++;
			ret++;
		}
		
		if ((b&8)!=0)
		{
			j++;
			ret++;
		}
		if ((b&16)!=0)
		{
			ret++;
		}
	}
	else
		ret=5;
	return(ret);
}

/*****************************************************************************
ApplyEffs
	Process envelope/panning.
	Kill silent, keyed off channels.
*****************************************************************************/
void ApplyEffs(void)
{
short envpan,envvol;
int t;
int pmsk;
short vlm;
short Pn;
u_char MaxChans;

 	ms->XMActiveVoices=0;
	MaxChans=ms->MaxChans;
 	for(t=0;t<mh->XMPSXChannels;t++)
 	{
		if (MaxChans==0)
			return;

 		pmsk=ms->PlayMask;
 		pmsk&=(1<<t);
 		if (pmsk!=0)
 		{
			MaxChans--;
 			XMC=&ms->XM_Chnl[t];

 			if(XMC->kick)							/* New note keyed on ? */
 			{
 				XMC->keyon=1;
 				XMC->fadevol=32767;
 				XMC->ChDead=0;
 			}
 			if (!XMC->ChDead)						/* Update envelope/pitch/vol */
 			{											/* If channel not dead */
 				ms->XMActiveVoices++;

				if(XMC->period<1)					/* Cap period */
					XMC->period=1;
				else if(XMC->period>32768)
					XMC->period=32768;

				if(XMC->envflg & EF_ON)
					envvol=ProcessEnvelope(256,XMC->keyon,XMC->sample);
				else
					envvol=256;
				if(XMC->panenvflg & EF_ON)
					envpan=ProcessPanEnvelope(128,XMC->keyon,XMC->sample);
				else
					envpan=128;

				vlm=((XMC->volume-64)*(64-XMC->UserVol))/64;
				vlm+=64;
				if (vlm<=64)
				{
					XMC->LVol=XMC->RVol=0;
				}
				else
				{
					vlm=((DoVol(XMC->fadevol,envvol,vlm-64)*ms->SongVolume)/128);
					vlm*=ms->MasterVolume;
					vlm/=128;

					Pn=XMC->panning; 
					Pn+=ms->UserPan; 
					if (Pn<0) 
						Pn=0; 
					else if (Pn>255)
						Pn=255; 
					Pn=DoPan(envpan,Pn); 

					XMC->LVol=(vlm*(255-Pn));	// /128;
					XMC->RVol=(vlm*Pn);			// /128;
				} 

				if ((!XMC->LVol)&&(!XMC->RVol)&&(!XMC->keyon)) 
				{ 
					StpCh(XMC->SPUChannel); 
					XMC->volume=64;
					XMC->ChDead=1;			/* Released & 0vol - so kill channel */
				} 
				else if(!XMC->keyon)		/* Key off? - do release on envelope */
				{ 
					XMC->fadevol-=XMC->keyoffspd;
					if(XMC->fadevol<0)
						 XMC->fadevol=0;
				}
			}
		}
	}
}

/*****************************************************************************
UpdateEffs
		Play any effects if required
*****************************************************************************/
void UpdateEffs(void) 
{ 
int t; 
int pmsk; 
u_char MaxChans; 
 
	MaxChans=ms->MaxChans; 
 	for(t=0;t<mh->XMPSXChannels;t++) 
 	{ 
		if (MaxChans==0) 
			return; 

 		XMC=&ms->XM_Chnl[t]; 
 		pmsk=ms->PlayMask; 
 		pmsk&=(1<<t); 
 		if (pmsk!=0) 
 		{
			MaxChans--;
 			if(!XMC->nothing)			/* Any note/vol/eff's to process? */
 			{
 				JPlayEffects();
 			} 
 		} 
 	} 
}
 
/*****************************************************************************
MonoCheck
		Creates a mono signal if you want to playback in mono mode..
*****************************************************************************/
void MonoCheck(void)
{
	if (MonoMode==XM_MONO)
	{
		DVL=(XMC->OldLVol+XMC->OldRVol)/2;
		DVR=DVL;
		return;
	}
	else
	{
		DVL=XMC->OldLVol;
		DVR=XMC->OldRVol;
	}
}

/*****************************************************************************
SetCore
		Sets SPU2 core number and channel depending on channel required
		Note: Use this call for all music calls (The ones that are in the
		interrupt)
*****************************************************************************/
int SetCore(int Chnl)
{
	if (Chnl<24)
	{
		CoreSD=SD_CORE_0;
		CoreNum=0;
		return (Chnl);
	}
	else
	{
		CoreSD=SD_CORE_1;
		CoreNum=1;
	}
	return (Chnl-24);
}

/*****************************************************************************
UserSetCore
		Sets SPU2 core number and channel depending on channel required
		Note: Use this call for any non music calls (So you don't wipe over
		the ones that are being used on an interrupt.)
*****************************************************************************/
int UserSetCore(int Chnl)
{
	if (Chnl<24)
	{
		UserCoreSD=SD_CORE_0;
		UserCoreNum=0;
		return (Chnl);
	}
	else
	{
		UserCoreSD=SD_CORE_1;
		UserCoreNum=1;
	}
	return (Chnl-24);
}

#define JPPIT 2682522
/*****************************************************************************
UpdateHardware
		Update SPU if needed
*****************************************************************************/
void UpdateHardware(void) 
{
int t;
int prd;
int SPUKeyOn[2];
int pmsk;
int Chnl; 
int Ch2;
u_char MaxChans;

	SPUKeyOn[0]=0;
	SPUKeyOn[1]=0;
	MaxChans=ms->MaxChans; 
	for(t=0;t<mh->XMPSXChannels;t++) 
	{ 
		if (MaxChans==0)
			break; 
 
		pmsk=ms->PlayMask; 
		pmsk&=(1<<t); 
		if (pmsk!=0) 
		{ 
			MaxChans--; 
			XMC=&ms->XM_Chnl[t]; 
			if(XMC->kick) 
			{
				Chnl=XMC->SPUChannel; 

				Ch2=SetCore(Chnl);

				sceSdSetSwitch (CoreSD | SD_S_KOFF, (1<<Ch2));
				prd=XMC->period; 
				if (ms->NotAmiga==0)
					prd=((JPPIT<<2)/prd)>>3;		/* Using AMIGA freq table */
				else
					prd=GetFreq2(prd);				/* Using LINEAR freq table */

				XMC->OldLVol=XMC->LVol; 
				XMC->OldRVol=XMC->RVol; 
				MonoCheck(); 
				SPUKeyOn[CoreNum]|=(1<<Ch2); 
				XMC->SPUPitch=prd; 
				XMC->OldPeriod=XMC->period; 

				PlaySFX(ms->VabID,Ch2,XMC->sample,prd,DVL,DVR);
			} 
		} 
	} 
		sceSdSetSwitch (SD_CORE_0 | SD_S_KON, SPUKeyOn[0]);
		sceSdSetSwitch (SD_CORE_1 | SD_S_KON, SPUKeyOn[1]);

	MaxChans=ms->MaxChans; 

	for(t=0;t<mh->XMPSXChannels;t++)
	{ 
		if (MaxChans==0) 
			return; 
		pmsk=ms->PlayMask;
		pmsk&=(1<<t); 
		if (pmsk!=0) 
		{ 
			MaxChans--; 
			XMC=&ms->XM_Chnl[t]; 
			if (!XMC->ChDead) 
			{ 
				Chnl=XMC->SPUChannel;
				Ch2=SetCore(Chnl);
				if (XMC->kick==0) 
				{
					if ((XMC->LVol!=XMC->OldLVol)||(XMC->RVol!=XMC->OldRVol)) 
					{
						if (XMC->OldLVol!=XMC->LVol) 
								XMC->OldLVol=IntVols(XMC->LVol,XMC->OldLVol); 

							if (XMC->OldRVol!=XMC->RVol) 
								XMC->OldRVol=IntVols(XMC->RVol,XMC->OldRVol); 
						MonoCheck(); 

						SetVol(CoreSD,Ch2,DVL,DVR);	// Update volume
					}
					if (XMC->period!=XMC->OldPeriod)
					{ 
						XMC->OldPeriod=XMC->period;
						prd=XMC->period; 
						if(ms->NotAmiga==0)
							prd=((JPPIT<<2)/prd)>>3;	/* Using AMIGA freq table */ 
						else 
							prd=GetFreq2(prd); 
						XMC->SPUPitch=prd; 
						SetFrq(CoreSD,Ch2,prd);			/* Update periods */ 
					} 
				} 
				else 
					XMC->kick=0; 
			} 
		} 
	} 
} 
 
/*****************************************************************************
IntVols
	Calculates point between Current volume and Required volume.
	Requires:
				Vol1		LVol
				Vol2		OldLVol
	Returns:
				value half way between two points.
*****************************************************************************/
int IntVols(int Vol1,int Vol2)
{
int a;
	a=abs(Vol2-Vol1);
	a+=2;
	a>>=2;

	if (Vol1<Vol2)
		return(Vol1+a);
	else
		return(Vol2+a);
}

/*****************************************************************************
GetFreq2
	Returns SPU pitch.
	Calculated from linear value
*****************************************************************************/
#define joc 8
int JPPer = 7350;
long GetFreq2(long period)
{
int okt;
long frequency;

	period=JPPer-period;
	okt=period/768;
	frequency=lintab[period%768];
	return(frequency>>(joc-okt));		/* Final SPU pitch */
}

/*****************************************************************************
ProcessEnvelope
	Updates amplitude envelope
*****************************************************************************/
short ProcessEnvelope(short v,u_char keyon,int JSmp)
{

u_char a2,b;
u_short p;

u_long* j;
short apos;
short aval;
short bpos;
short bval;

	a2=XMC->enva;
	b=XMC->envb;
	p=XMC->envp;

	/* compute the envelope value between points a and b */

	j=mh->JAP_InstrumentOffset[JSmp];
	aval=(getSWord((u_char*)j+129+2+(a2*4)))<<2;
	bval=(getSWord((u_char*)j+129+2+(b*4)))<<2;
	apos=getSWord((u_char*)j+129+(a2*4));
	bpos=getSWord((u_char*)j+129+(b*4));

	v=Interpolate(p,apos,bpos,aval,bval);

		/* Should we sustain? (sustain flag on, key-on, point a is the sustain
		   point, and the pointer is exactly on point a) */

	if((XMC->envflg & EF_SUSTAIN) && keyon && a2==XMC->envsus && p==apos)
	{
		/* do nothing */
	}
	else
	{
		p++;							/* don't sustain, so increase pointer. */
		if(p >= bpos)				/* pointer reached point b? */
		{
			a2=b; b++;				/* shift points a and b */
			if(XMC->envflg & EF_LOOP)
			{
				if(b > XMC->envend)
				{
					a2=XMC->envbeg;
					b=a2+1;
					p=getSWord((u_char*)j+129+(a2*4));
				}
			}
			else
			{
				if(b >= XMC->envpts)
				{
					b--;
					p--;
				}
			}
		}
	}
	XMC->enva=a2;
	XMC->envb=b;
	XMC->envp=p;

	return v;
}

/*****************************************************************************
ProcessPanEnvelope
	Updates panning envelope
*****************************************************************************/
short ProcessPanEnvelope(short v,u_char keyon,int JSmp)
{

u_char a2,b;
u_short p;

u_long* j;
short apos;
short aval;
short bpos;
short bval;

				
	a2=XMC->panenva;
	b=XMC->panenvb;
	p=XMC->panenvp;

	/* compute the panenvelope value between points a and b */

	j=mh->JAP_InstrumentOffset[JSmp];
	aval=(getSWord((u_char*)j+177+2+(a2*4)))<<2;
	bval=(getSWord((u_char*)j+177+2+(b*4)))<<2;
	apos=getSWord((u_char*)j+177+(a2*4));
	bpos=getSWord((u_char*)j+177+(b*4));

	v=Interpolate(p,apos,bpos,aval,bval);

		/* Should we sustain? (sustain flag on, key-on, point a is the sustain
		   point, and the pointer is exactly on point a) */

	if((XMC->panenvflg & EF_SUSTAIN) && keyon && a2==XMC->panenvsus && p==apos)
	{
		/* do nothing */
	}
	else
	{
		p++;							/* don't sustain, so increase pointer. */

		if(p >= bpos)				/* pointer reached point b? */
		{
			a2=b; b++;				/* shift points a and b */
			if(XMC->panenvflg & EF_LOOP)
			{
				if(b > XMC->panenvend)
				{
					a2=XMC->panenvbeg;
					b=a2+1;
					p=getSWord((u_char*)j+177+(a2*4));
				}
			}
			else
			{
				if(b >= XMC->panenvpts)
				{
						b--;
						p--;
				}
			}
		}
	}
	XMC->panenva=a2;
	XMC->panenvb=b;
	XMC->panenvp=p;
	return v;
}

/*****************************************************************************
XM_SetSongPos
	User Call. Set song position to requested value.
*****************************************************************************/
void XM_SetSongPos(int Song_ID,u_short pos)
{
int t;
	if (Song_ID>31)
		return;
	if (XMSongIDs[Song_ID]==-1)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	mhu=(XMHEADER*)(XM_HeaderAddress[mu->HeaderNum]);

	if (pos>=mhu->songlength)
		return;

	JP_Do_Nothing=1;	// Stops interrupt from updating player..

	for(t=0;t<mhu->XMChannels;t++)
	{
		XMCU=&mu->XM_Chnl[t];
//		XMCU->keyon=0;
		XMCU->tmpvolume=64;
		XMCU->LVol=0;
		XMCU->RVol=0;
	}

	mu->posjmp=1;
	mu->patbrk=0;
	mu->SongPos=pos;
	mu->vbtick=mu->SongSpeed;
	JP_Do_Nothing=0;
}

/*****************************************************************************
XM_SetChVolume
	User Call. Set channel to a volume
	0-loud
	64-quiet.
*****************************************************************************/
void XM_SetChVolume(int Song_ID,int Channel,int Volume)
{

	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	XMCU=&mu->XM_Chnl[Channel];
	if (Volume>64)
		Volume=64;
	Volume=64-Volume;
	XMCU->UserVol=Volume;
}

/*****************************************************************************
XM_GetChVolume
	User Call. Returns channel user volume

*****************************************************************************/
int XM_GetChVolume(int Song_ID,int Channel)
{
int vlm;
	if (XMSongIDs[Song_ID]==-1)
		return(-1);
	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	XMCU=&mu->XM_Chnl[Channel];
	vlm=(XMCU->LVol+XMCU->RVol)>>8;

	vlm*=XMCU->kick2;
	XMCU->kick2=XMCU->kick;
	return (vlm);
}
 
/*****************************************************************************
PlaySFX
	Requires:
				VBID		ID of sample bank
				Channel		SPU Channel (0-23)
				Inst		Sample to play
				Pitch		Pitch to play sample at
				LV			Left Volume
				RV			Right Volume
****************************************************************************/
void PlaySFX(int VBID,int Channel,int Inst,int Pitch,int LV,int RV)
{
u_int addr;
//u_int addr2;

	sceSdSetAddr (CoreSD | Channel<<1 | SD_VA_SSA, xm_l_vag_spu_addr[VBID][Inst]+XMC->SOffset);
	SetFrq(CoreSD,Channel,Pitch);
	SetVol(CoreSD,Channel,LV,RV);


	if ((XMC->LoopFlag!=0)&&(XMC->SOffset!=0))
	{
		addr=XMC->LoopStart;	// Set approx loop start position if sample is looping and is not getting played
//		addr*=7;				// from the begining (just in case it start address if after the loop start marker)
//		addr/=8;
		addr*=2;
		addr/=7;
		addr&=(~31);
//		addr2=addr;

		addr+=xm_l_vag_spu_addr[VBID][Inst];
//		printf("addr = %x, next = %x,loop=%x(%x) original %x\n",xm_l_vag_spu_addr[VBID][Inst],xm_l_vag_spu_addr[VBID][Inst+1],addr,addr2,XMC->LoopStart);
		sceSdSetAddr (CoreSD | Channel<<1 | SD_VA_LSAX, addr);
	}

}

/*****************************************************************************
InitSPUChannel
	Sets up ADSR for selected channel
	Called only when Init_XM
*****************************************************************************/
void InitSPUChannel(int Channel)
{
int Ch2;
	Ch2=SetCore(Channel);
	sceSdSetParam (CoreSD | Ch2<<1 | SD_VP_ADSR1, 0x060f);	//0x4fff);
	sceSdSetParam (CoreSD | Ch2<<1 | SD_VP_ADSR2, 0x1fc1);	//0x1fc0);
}

/*****************************************************************************
CurrentKeyStat
	Stops processing of any dead channels
*****************************************************************************/
void CurrentKeyStat(void) 
{
int t; 
int z;
int ch2;
	for (t=0;t<mh->XMPSXChannels;t++) 
	{ 
		XMC=&ms->XM_Chnl[t];
		if (XMC->ChDead!=1)
		{
			z=XMC->SPUChannel;
			ch2=SetCore(z);
			z=sceSdGetParam(CoreSD|ch2<<1|SD_VP_ENVX);
			if (z==0)		//Key'ed off?
			{ 
				XMC->ChDead=1; 
				XMC->nothing=1;

			}
		} 
	} 
}
 
/*****************************************************************************
StpCh
	Turns off channel

	Requires	Channel		SPU Channel number 0-47
*****************************************************************************/
void StpCh(int Channel) 
{ 
int ch2;
	ch2=SetCore(Channel);
	sceSdSetSwitch (CoreSD | SD_S_KOFF, (1<<ch2));
	sceSdSetParam (CoreSD | ch2<<1 | SD_VP_VOLL, 0);
	sceSdSetParam (CoreSD | ch2<<1 | SD_VP_VOLR, 0);
}

/*****************************************************************************
SetVol
	Sets volume for channel

	Requires	core		0 or 1
				Channel		SPU Channel number 0-23
				LVol		Left Volume
				RVol		Right Volume
*****************************************************************************/
void SetVol(int core,int Channel,int LVol,int RVol)
{
	sceSdSetParam (core | Channel<<1 | SD_VP_VOLL, LVol);
	sceSdSetParam (core | Channel<<1 | SD_VP_VOLR, RVol);
}

/*****************************************************************************
SetFrq
	Sets frequency (pitch) for channel

	Requires	core		Core to use (0 or 1)
				Channel		SPU Channel number 0-23
				Pitch		SPU Pitch (currently in 44.1khz format)
*****************************************************************************/
void SetFrq(int core,int Channel,int Pitch) 
{ 
	Pitch*=44100;		// change pitch to 48khz format
	Pitch/=48000;
	sceSdSetParam (core | Channel<<1 | SD_VP_PITCH, Pitch);
}
 
/*****************************************************************************
SilenceXM
	Silence any channel used for Song_ID
*****************************************************************************/
void SilenceXM(int Song_ID)
{
int t,i,c,c2;

int pmsk;


	mu=(XMSONG*)(XM_SngAddress[Song_ID]); 
	mhu=(XMHEADER*)(XM_HeaderAddress[mu->HeaderNum]); 
	i=0; 
	for(t=0;t<mhu->XMPSXChannels;t++) 
	{ 
		pmsk=mu->PlayMask; 
		pmsk&=(1<<t); 
		if (pmsk!=0) 
		{
			XMC=&mu->XM_Chnl[t]; 
			c=XMC->SPUChannel; 
			c2=UserSetCore(c);

			i|=(1<<c2);
			SetVol(UserCoreSD,c2,0,0); 
			sceSdSetSwitch (UserCoreSD | SD_S_KOFF, (1<<c2));
		} 
	} 
}
 
/*****************************************************************************
XM_Pause
	User Call. Pauses Song_ID
	(by setting pitch to 0 - not by keying off..)
*****************************************************************************/
void XM_Pause(int Song_ID)
{
int t;
int c,c2;

	if (Song_ID>31)
		return;


	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	mhu=(XMHEADER*)(XM_HeaderAddress[mu->HeaderNum]); 
	if (mu->XMPlay==XM_PLAYING) 
	{
		mu->XMPlay=XM_PAUSED;
		for(t=0;t<mhu->XMPSXChannels;t++) 
		{
			XMCU=&mu->XM_Chnl[t]; 
			c=XMCU->SPUChannel;
			c2=UserSetCore(c);
			SetFrq(UserCoreSD,c2,0); 
		} 
	} 
} 
 
void XM_PauseAll(void)
{
int i;
	for (i=0;i<32;i++)
	{
		XM_Pause(i);
	}
}

void XM_RestartAll(void)
{
int i;
	for (i=0;i<32;i++)
	{
		XM_Restart(i);
	}
}

/*****************************************************************************
XM_Restart
	User Call. Restarts Song_ID
	(returns the channels pitch to the original value)
*****************************************************************************/
void XM_Restart(int Song_ID)
{
int t,c,c2;

	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	mhu=(XMHEADER*)(XM_HeaderAddress[mu->HeaderNum]);
	if (mu->XMPlay==XM_PAUSED)
	{
		mu->XMPlay=XM_PLAYING;
		for(t=0;t<mhu->XMPSXChannels;t++)
		{
			XMCU=&mu->XM_Chnl[t];
			if(!XMCU->ChDead)
			{
				c=XMCU->SPUChannel; 
				c2=UserSetCore(c);
				SetFrq(UserCoreSD,c2,XMCU->SPUPitch);
			}
		}
	}
}

/*****************************************************************************
XM_PlayNext
	User Call. Set a song pos to jump to when play reaches the end of the
	current pattern.

	Requires:
				Song ID
				SongPos in pattern list to play next. or -1 to cancel
*****************************************************************************/
void XM_PlayNext(int Song_ID,short SongPos)
{
	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	mu->PlayNext=SongPos;
}

/*****************************************************************************
XM_SetMasterVol
	User Call. Set Master Volume from 0-128;

	Requires:
				Song ID
				Volume;
*****************************************************************************/
void XM_SetMasterVol(int Song_ID,u_char Vol)
{
	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);

	if (Vol<=128)
	{
		mu->MasterVolume=Vol;
	}
}

/*****************************************************************************
XM_SetMasterPan
	User Call. Set Master Panning from -127 to 127;

	Requires:
				Song ID
				PanPos;
*****************************************************************************/
void XM_SetMasterPan(int Song_ID,short Pan)
{
	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);

	mu->UserPan=Pan;
}

/*****************************************************************************
XM_Quit
	Stop playing song. Kill any SPU Channels
****************************************************************************/
void XM_Quit(int SongID)
{
	XM_PlayStop(SongID);
}

/*****************************************************************************
XM_PlaySample
	Play a sample from one of the SampleBank files.

	Requires	vid			sample bank ID
				slot		sample number
				channel		SPU Channel to play sample on (0-47)
				voll		Left Volume
				volr		Right Volume
				pitch		SPU Pitch
****************************************************************************/
void XM_PlaySample(int vid,int slot,int channel,int voll,int volr,int pitch)
{
int ch2;
int core;
u_int addr;

	if (channel<24)
	{
		core=SD_CORE_0;		// just so we are not wiping over the values used by the code on interrupt.
		ch2=channel;
	}
	else
	{
		core=SD_CORE_1;
		ch2=channel-24;
	}

	addr=xm_l_vag_spu_addr[vid][slot];
	sceSdSetParam (core | ch2<<1 | SD_VP_ADSR1, 0x000f);	//0x4fff);
	sceSdSetParam (core | ch2<<1 | SD_VP_ADSR2, 0x1fc0);	//0x1fc0);
	sceSdSetParam (core | ch2<<1 | SD_VP_VOLL, voll);
	sceSdSetParam (core | ch2<<1 | SD_VP_VOLR, volr);
	sceSdSetParam (core | ch2<<1 | SD_VP_PITCH, pitch);
	sceSdSetAddr (core | ch2<<1 | SD_VA_SSA, addr);
	sceSdSetSwitch (core | SD_S_KON, (1<<ch2));

} 

/*****************************************************************************
XM_StopSample
	Stop a sample from one of the SampleBank files.
****************************************************************************/
void XM_StopSample(int channel)
{
int ch2;
int core;

int i;

	for (i=0;i<1000;i++)
	{
	}
	if (channel<24)
	{
		core=SD_CORE_0;		// just so we are not wiping over the values used by the code on interrupt.
		ch2=channel;
	}
	else
	{
		core=SD_CORE_1;
		ch2=channel-24;
	}

	sceSdSetSwitch (core | SD_S_KOFF, (1<<ch2));
}

/*****************************************************************************
XM_SetVAGAddress
	Set address of sample in SPU RAM into Samplebank slot.
****************************************************************************/
void XM_SetVAGAddress(int VabID,int slot,int addr)
{
	xm_l_vag_spu_addr[VabID][slot] = addr;
}

/*****************************************************************************
XM_GetSongSize
	Returns the size of the XMSONG Structure, so this memory can be allocated
	Up to 32 XM SONG's can be played at once, from up to 8 XM/Sample files..
****************************************************************************/
int XM_GetSongSize(void)
{
	return(sizeof(XMSONG));
}

/*****************************************************************************
XM_SetSongAddress
	Set address of SONG buffer. This buffer holds all the info for a song,
	such as play position, bpm, speed etc..
****************************************************************************/
void XM_SetSongAddress(u_char *Address,int song)
{
	mu=((XMSONG*)(Address));
	XMSongIDs[song]=-1;
	mu->XMPlay=XM_NOT_INIT;				// Must set this before the address is stored!!
 	XM_SngAddress[song]=Address;
}


/*****************************************************************************
XM_GetFileHeaderSize
	Returns size of XMHEADER structure. This is then allocated and it is where
	data for each XM file is copied into (such as original bpm, channels used)
	Up to 8 of these can be opened at one time
****************************************************************************/
int XM_GetFileHeaderSize(void)
{
	return(sizeof(XMHEADER));
}

/*****************************************************************************
XM_SetFileHeaderAddress
	Set address of allocated memory for XMHEADER
****************************************************************************/
void XM_SetFileHeaderAddress(u_char *Address,int id)
{
  	XM_HeaderAddress[id]=Address;
}

/*****************************************************************************
XM_SetBPM
	Set new BPM for song
****************************************************************************/
void XM_SetBPM(int Song_ID,u_short BPM)
{
	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	if (mu->XMPlay==XM_NOT_INIT)
		return;
 	mu->SongBPM=BPM;
}

/*****************************************************************************
XM_SetSpeed
	Set new Speed for song
****************************************************************************/
void XM_SetSpeed(int Song_ID,u_short Speed)
{
	if (XMSongIDs[Song_ID]==-1)
		return;

	if (XM_SngAddress[Song_ID]==0)
		return;

	mu=(XMSONG*)(XM_SngAddress[Song_ID]);
	if (mu->XMPlay==XM_NOT_INIT)
		return;
 	mu->SongSpeed=Speed;
}

/*****************************************************************************
XM_SetMono
	Set output for all songs to MONO
****************************************************************************/
void XM_SetMono(void)
{
	MonoMode=XM_MONO;
}

/*****************************************************************************
XM_SetStereo
	Set output for all songs to STEREO
****************************************************************************/
void XM_SetStereo(void)
{
	MonoMode=XM_STEREO;
}

 
static unsigned int rpc_arg [512];	// Size of Buffer sent from EE....
volatile int ret[128];	//= 0;		// Buffer sent from IOP to EE...

/**********************************************************************************
xmplay_loop
	Kicks off everything..
***********************************************************************************/
int xmplay_loop (void)
{
sceSifQueueData qd;
sceSifServeData sd;

    CpuEnableIntr ();
    EnableIntr (INUM_DMA_4);
    EnableIntr (INUM_DMA_7);
    EnableIntr (INUM_SPU);
    
    sceSifInitRpc (0);
    sceSifSetRpcQueue (&qd, GetThreadId ());
    sceSifRegisterRpc (&sd, SOUND_DEV, ProcessEECommand, (void*)rpc_arg, NULL, NULL, &qd);
    sceSifRpcLoop (&qd);

    return 0;
}

/**********************************************************************************
ProcessEECommand
	Using SIF DMA, the EE sends control commands to this routine.
***********************************************************************************/
static void* ProcessEECommand (unsigned int command, void *data_, int size)
{ 
int ch;
int          data  = *((         int *) data_);

    ch = command & SOUND_CH_MASK;
    switch (command & SOUND_COMMAND_MASK)
    {
	case SOUND_SETPARAMS:
		AdpcmSetParam ( ch, (short *)data_);
		break;
    default:
		ERROR (("Driver error: unknown command %d \n", data));
	break;
    }
	SetReturnBuffer();
    return (void*)(&ret[0]);
}

int Buffer64K=0;
int VBID;
/**********************************************************************************
AdpcmSetParam
	Decodes the command/data packets sent from the EE
***********************************************************************************/
void AdpcmSetParam(int ch, short *jdata)
{
int b,i,f,j;
u_short dt[128];
int length;

	b=*jdata;
	jdata++;

	for (i=0;i<b;i++)
	{
		f=*jdata;			// Function;
		jdata++;
		length=*jdata;		// amount of data for required for function.
		jdata++;

		for (j=0;j<length;j++)
		{
			dt[j]=*jdata;
			jdata++;
		}

		if (f==XMI_INIT_SPU)
			EE_SdInit(0,0);		// data means nothing
		else if (f==XMI_SetTransAddress)
			EE_SetSPUBaseAddress(dt[0],dt[1]);
		else if (f==XMI_CreateTransBuff)
			EE_CreateTransferBuffer();
		else if (f==XMI_FreeTransBuff)
			EE_FreeTransferBuffer();
		else if (f==XMI_SetSampleOffsets)
			EE_CreateSampleOffsetData(dt[0],dt[1],dt[2],dt[3]);
		else if (f==XMI_AllocXMBuffer)
			EE_AllocXMBuffer(dt[0]);
		else if (f==XMI_AllocXMSong)
			EE_AllocXMSong(dt[0]);
		else if (f==XMI_FreeXMPatternRAM)
			EE_FreeXMPatternRAM(dt[0]);
		else if (f==XMI_AllocXMPatternRAM)
			EE_AllocateXMPatternRAM(dt[0],dt[1],dt[2]);
		else if (f==XMI_InitXMData)
			EE_InitXMData(dt[0],dt[1]);
		else if (f==XMI_PlayXM)
			EE_XMPlay(dt[0],dt[1],dt[2],dt[3],dt[4],dt[5],dt[6],dt[7]);
		else if (f==XMI_XMInit)
			EE_XMInit();
		else if (f==XMI_TransferSample)
			EE_TransferSampleData(dt[0],dt[1]);
		else if (f==XMI_Pause)
			EE_Pause(dt[0]);
		else if (f==XMI_Resume)
			EE_Resume(dt[0]);
		else if (f==XMI_PauseAll)
			EE_PauseAll();
		else if (f==XMI_ResumeAll)
			EE_ResumeAll();
		else if (f==XMI_FreeXMBuffer)
			EE_FreeXMBuffer(dt[0]);
		else if (f==XMI_XMExit)
			EE_XMExit();
		else if (f==XMI_XMQuit)
			EE_XMQuit(dt[0]);
		else if (f==XMI_SetSongPos)
			EE_SetSongPos(dt[0],dt[1]);
		else if (f==XMI_SetChannelVolume)
			EE_SetChannelVolume(dt[0],dt[1],dt[2]);
		else if (f==XMI_FreeXMSong)
			EE_FreeXMSong(dt[0]);
		else if (f==XMI_SetMasterVolume)
			EE_SetMasterVol(dt[0],dt[1]);
		else if (f==XMI_SetMasterPan)
			EE_SetMasterPan(dt[0],dt[1]);
		else if (f==XMI_GetSongFeedback)
			EE_GetSongFeedback(dt[0]);
		else if (f==XMI_SetBPM)
			EE_SetBPM(dt[0],dt[1]);
		else if (f==XMI_SetSpeed)
			EE_SetSpeed(dt[0],dt[1]);
		else if (f==XMI_SetMono)
			XM_SetMono();
		else if (f==XMI_SetStereo)
			XM_SetStereo();
		else if (f==XMI_PlaySample)
			XM_PlaySample(dt[0],dt[1],dt[2],dt[3],dt[4],dt[5]);
		else if (f==XMI_StopSample)
			XM_StopSample(dt[0]);
		else if (f==XMI_PlayNextPos)
			XM_PlayNext(dt[0],dt[1]);
	}
}

/**********************************************************************************
EE_SdInit
	Init SPU2 and set Master Volume
***********************************************************************************/
void EE_SdInit (int no_ch, int status)
{
    sceSdInit (0);

    //    Disk media: DVD
    // Output format: PCM
    //    Copy guard: normal (one generation recordable / default)
    sceSdSetCoreAttr (SD_C_SPDIF_MODE,
    				 (SD_SPDIF_MEDIA_DVD | SD_SPDIF_OUT_PCM | SD_SPDIF_COPY_NORMAL));

    sceSdSetParam (SD_CORE_0 | SD_P_MVOLL, 0x3fff);
    sceSdSetParam (SD_CORE_0 | SD_P_MVOLR, 0x3fff);
    sceSdSetParam (SD_CORE_1 | SD_P_MVOLL, 0x3fff);
    sceSdSetParam (SD_CORE_1 | SD_P_MVOLR, 0x3fff);

    if (gSem == 0)
    {
		struct SemaParam sem;
		sem.initCount = 0;
		sem.maxCount = 1;
		sem.attr = AT_THFIFO;
		gSem = CreateSema (&sem);			// Create Sema

    }

	if( gThid == 0 )
	{
		gThid = makeMyThread();				// Create Thread
		StartThread( gThid, (u_long)NULL );
	}

	setTimer(&timer);		// Setup IRQ interrupt
	startTimer(&timer);		// Start IRQ interrupt (calls TimerFunc)
    return;
}

/**********************************************************************************
setTimer - IRQ Interrupt
***********************************************************************************/
void setTimer( TimerCtx* timer )
{
	// --- use H-Sync timer
	timer->timerID = AllocHardTimer(TC_HLINE , 16, 1);			// CHANGE TO RUN ON VSYNC - NOT HSYNC..
	RegisterIntrHandler(GetHardTimerIntrCode(timer->timerID),
			HTYPE_C, (void*)TimerFunc, timer );
	SetTimerCompare(timer->timerID, 154);	//154//33*5);	// 70 fps (66 = 1/240th of a second, so 198=1/70th)
	SetTimerMode( timer->timerID , tEXTC_1 | tGATF_0 | tZRET_1 | tREPT_1 | tCMP_1);
}

/**********************************************************************************
startTimer - IRQ Interrupt
***********************************************************************************/
void startTimer( TimerCtx* timer )
{
	EnableIntr(GetHardTimerIntrCode(timer->timerID));
}

/**********************************************************************************
IRQ Interrupt - called, say, once per frame
***********************************************************************************/
int TimerFunc( TimerCtx* timer )
{
	int m ;
	m = iGetTimerStatus(timer->timerID);

	if ( m & tEQUF_1 )
	{
		VCount++;
		iSignalSema(gSem);	// Signal StreamADPCM to be called
	}
	return 1;
}


/**************************************************************************/
/* Create thread that gets called by callback function					  */
/**************************************************************************/
static int makeMyThread( void )
{
    struct ThreadParam param;
    int	thid;

    param.attr         = TH_C;
    param.entry        = UpdateXM;		// main sample load/convert/play routine
    param.initPriority = BASE_priority-3;
    param.stackSize    = 0x1000;
    param.option = 0;
	thid = CreateThread(&param);
	return thid;
}

int UpdateXM(int status)
{
	while(1)
	{
		WaitSema(gSem);
		UpdateXMData();
	}
	return(0);
}

/**********************************************************************************
EE_AllocXMBuffer
	Allocate XMBuffer (for up to 8 XM Files)
***********************************************************************************/
void EE_AllocXMBuffer(int id)
{
int FileHSize,FHAddr;

	if ((id<8)&&(XM_HeaderAddress[id]==0))
	{
		FileHSize=XM_GetFileHeaderSize();
		FHAddr = (int) AllocSysMemory (0, FileHSize, NULL);
		XM_SetFileHeaderAddress((void*)FHAddr,id);
	}
}

/**********************************************************************************
EE_FreeXMBuffer
	Free XMBuffer (for up to 8 XM Files)
***********************************************************************************/
void EE_FreeXMBuffer(int xmd)
{
	if ((xmd<8)&&(XM_HeaderAddress[xmd]!=0))
	{
	 	FreeSysMemory( (void*)XM_HeaderAddress[xmd]);
		XM_HeaderAddress[xmd]=0;
	}
}

u_int XMPatAddr[8]={0,0,0,0,0,0,0,0};
int LastPatAlloc=0;
int Song_ID=-1;	// (Pass this back to EE - it contians the last ID returned by XM_Init.)
int testcounter=0;
/**********************************************************************************
EE_AllocXMSong
	Alloc buffer for XM Song (up to 32 songs can be alloced/played)
***********************************************************************************/
void EE_AllocXMSong(int Song)
{
int SongSize;
u_int addr;
	SongSize=XM_GetSongSize();
	if ((Song<32)&&(XM_SngAddress[Song]==0))
	{
		addr = (int) AllocSysMemory (0, SongSize, NULL);
		XM_SetSongAddress((void*)addr,Song);
	}
}

/**********************************************************************************
EE_FreeXMSong
	Free XM song buffer
***********************************************************************************/
void EE_FreeXMSong(int song)
{
	if ((song<32)&&(XM_SngAddress[song]!=0))
	{
		FreeSysMemory( (void*)XM_SngAddress[song]);
		XM_SngAddress[song]=0;
		XMSongIDs[song]=-1;
	}
}

/**********************************************************************************
EE_FreeXMPatternRAM
	Free buffer that holds the XM pattern data (the actual XM file)
***********************************************************************************/
void EE_FreeXMPatternRAM(int xmd)
{
	if ((xmd<8)&&(XMPatAddr[xmd]!=0))
	{
		FreeSysMemory( (void*)XMPatAddr[xmd]);
		XMPatAddr[xmd]=0;
	}
}

/**********************************************************************************
EE_AllocateXMPatternRAM
	Allocate memory for buffer that holds the XM pattern data (the actual XM file)
***********************************************************************************/
void EE_AllocateXMPatternRAM(unsigned int s1,unsigned int s2,int xmd)
{
unsigned int XMFileSize;

	XMFileSize=s1<<16;
	XMFileSize+=s2;

	if ((xmd<8)&&(XMPatAddr[xmd]==0))
	{
		XMPatAddr[xmd] = (int)AllocSysMemory( 0, XMFileSize, NULL );		// Allocate buffer for loading..
	}
 	LastPatAlloc=XMPatAddr[xmd];
}

/**********************************************************************************
EE_InitXMData
	Initialise the XM file (copies certain bits of the XM File into the XMBUFFER)
***********************************************************************************/
void EE_InitXMData(int id, int pan)
{
int chans;
int address;

	if (id<8)
	{
		if (XM_HeaderAddress[id]!=0)
		{
			address=XMPatAddr[id];
			if (address!=0)
				chans=InitXMData((void*)address,id,pan);	//XM_ID,XM_UseXMPanning); 	/* Address, XM_ID, PanType */
		}
	}
}

/**********************************************************************************
EE_XMPlay
	At last!! Actually play some music!!
***********************************************************************************/
void EE_XMPlay(int V_ID,int X_ID,int S_ID,int firstchan,int loop,u_int bitmask,u_int bitmask2,int offset)
{
unsigned int bm;
u_int a;
	bm=bitmask<<16;
	bm+=bitmask2;

 	a=XM_SngAddress[S_ID];
	if (a==0)
		return;

	Song_ID=XM_Init
				(V_ID,			// Sample Bank Handle
				X_ID,			// XM ID 
			    S_ID,			// SongID
				firstchan,	   	// First Channel To Playback On
				loop,			// Looping or once off tune? (XM_Loop, XM_NoLoop)
				bm,		// Play bitmask (set to 1 to allow channel to playback)
				offset);		// Start offset (song position)
}

/**********************************************************************************
EE_XMInit
	Do this once at the very start. Just initialises a few params..
***********************************************************************************/
void EE_XMInit(void)
{
	XM_OnceOffInit();
	XM_SetStereo();				/* or XM_SetMono for mono output */ 
}

/**********************************************************************************
EE_CreateSampleOffsetData
	Set address of sample n in sample bank y
***********************************************************************************/
void EE_CreateSampleOffsetData(int V,int slot,u_short ad1,u_short ad2)
{
unsigned int address;

	address=(ad1<<16)+ad2;
	if (address!=0)
		XM_SetVAGAddress(V,slot,address);
}
	
/**********************************************************************************
EE_CreateTransferBuffer
	Creates a 64K buffer that is then used as a transfer point between EE and SPU
	RAM
***********************************************************************************/
void EE_CreateTransferBuffer(void)
{
	if (Buffer64K==0)
	{
		Buffer64K = (int)AllocSysMemory( 0, (64*1024), NULL );		// Allocate buffer for loading..
	}
}

/**********************************************************************************
EE_FreeTransferBuffer
	Free's the transfer buffer used for sample transfer between EE and SPU
***********************************************************************************/
void EE_FreeTransferBuffer(void)
{
	if (Buffer64K!=0)
	{
		FreeSysMemory( (void*)Buffer64K);
		Buffer64K = 0;
	}
}

/**********************************************************************************
EE_SetSPUBaseAddres
	Set the address that samples will start to be transferred to. This value is
	also sent back to the EE.
***********************************************************************************/
void EE_SetSPUBaseAddress(unsigned int address,unsigned int address2)
{
unsigned int addr;

	addr=address<<16;
	addr+=address2;
	VAG_Addr=addr;
}


/***************************************************************************** 
TransferSampleData
	Transfers a block of sample data to SPU RAM.
	IOP data buffer is a max of 64K in size.
*****************************************************************************/
void EE_TransferSampleData(unsigned int size,unsigned int sizeb)
{
unsigned int size2;

	size2=size<<16;
	size2+=sizeb;
	if (size2==0)
		return;

 	sceSdVoiceTrans (0, (SD_TRANS_MODE_WRITE | SD_TRANS_BY_DMA),
 	(unsigned char *) Buffer64K,
 	(unsigned char *) VAG_Addr,size2);
	sceSdVoiceTransStatus(0,SD_TRANS_STATUS_WAIT);
	VAG_Addr+=size2;
}

/**********************************************************************************
EE_Pause
	Pause a song
***********************************************************************************/
void EE_Pause(int songid)
{
	XM_Pause(songid);
}

/**********************************************************************************
EE_Resume
	Resume a song
***********************************************************************************/
void EE_Resume(int songid)
{
	XM_Restart(songid);
}

/**********************************************************************************
EE_PauseAll
	Pause all playing songs
***********************************************************************************/
void EE_PauseAll(void)
{
	XM_PauseAll();
}

/**********************************************************************************
EE_ResumeAll
	Resume all playing songs
***********************************************************************************/
void EE_ResumeAll(void)
{
	XM_RestartAll();
}

/**********************************************************************************
EE_XMQuit
	Stop playing a song + shut down SPU channels
***********************************************************************************/
void EE_XMQuit(int songid)
{
	if (songid<32)
		XM_Quit(songid);
}

/**********************************************************************************
EE_XMExit
	Stop playing all songs and shut down SPU channels. Call at end of program
***********************************************************************************/
void EE_XMExit(void)
{
	XM_Exit();
}

/**********************************************************************************
EE_SetSongPos
	Set position through song
***********************************************************************************/
void EE_SetSongPos(int song,int pos)
{
	if ((song<32)&&(XM_SngAddress[song]!=0))
		XM_SetSongPos(song,pos);
}

/**********************************************************************************
EE_SetChannelVolume
	Set a channel volume (0-64) for a channel within a song
***********************************************************************************/
void EE_SetChannelVolume(int songid,int channel,int volume)
{
	XM_SetChVolume(songid,channel,volume);	// 0-64
}

/**********************************************************************************
EE_SetMasterVolume
	Set master volume for song (0-128)
***********************************************************************************/
void EE_SetMasterVol(int song,int vol)	// 0-128
{
	XM_SetMasterVol(song,vol);
}

/**********************************************************************************
EE_SetMasterPan
	Set master pan position for song (-127 = far left, +127 = far right)
***********************************************************************************/
void EE_SetMasterPan(int song,int pan)
{
	XM_SetMasterPan(song,pan);	//-127 to 127
}

/**********************************************************************************
EE_GetSongFeedback
	Files return buffer with info from selected song id
***********************************************************************************/
void EE_GetSongFeedback(int songid)
{
	EE_SongFeedbackID=songid;
}

/**********************************************************************************
EE_SetBPM
	Set BPM for song
***********************************************************************************/
void EE_SetBPM(int song,int bpm)
{
	XM_SetBPM(song,bpm);
}

/**********************************************************************************
EE_SetSpeed
	Set Speed for song
***********************************************************************************/
void EE_SetSpeed(int song,int bpm)
{
	XM_SetSpeed(song,bpm);
}

/**********************************************************************************
SetReturnBuffer
	Fill return buffer with lots of useful info.
***********************************************************************************/
void SetReturnBuffer(void)
{
int i;
int t;
	ret[0]=VAG_Addr;			// Current SPU RAM position for transfering samples to
	ret[1]=VBID;				// Last VAB ID created
	ret[2]=LastPatAlloc;		// Last address allocated for storing XM pattern data
	ret[3]=testcounter;			// A counter - use it if you like...
	ret[4]=Buffer64K;			// Address of 64K buffer
	testcounter++;

#if 1
	if (XM_SngAddress[EE_SongFeedbackID]!=0)
	{
		mu=(XMSONG*)(XM_SngAddress[EE_SongFeedbackID]);
		t=mu->XMPlay;
		if (t!=XM_NOT_INIT)
		{
			if (XMSongIDs[EE_SongFeedbackID]!=-1)
			{
				mhu=(XMHEADER*)(XM_HeaderAddress[mu->HeaderNum]);
				ret[6]=mu->CurrentPattern;
				ret[7]=mu->SongPos;
				ret[8]=mu->SongBPM;
				ret[9]=mu->SongSpeed;
				ret[10]=mu->XMPlay;		// Status
				ret[11]=mu->PatternPos;	
				ret[12]=mhu->songlength;
				ret[13]=mhu->XMChannels;

				for (i=0;i<32;i+=2)			// Store 32 volumes (if a new keyon has happened) as 16 words
				{
					t=XM_GetChVolume(EE_SongFeedbackID,i);
					t<<=8;
					t+=XM_GetChVolume(EE_SongFeedbackID,i+1);
					ret[14+(i/2)]=t;
				}
			}
  		}
		else
		{
			ret[10]=XM_NOT_INIT;
		}
	}
	else
	{
		ret[10]=XM_NO_SONG_ALLOCATED;
		for (i=0;i<16;i++)
		{
			ret[14+i]=0;	// Set volumes to 0..
		}
	}
#endif
}


